import numpy as np
import pandas as pd
from pyspark.sql import functions as F
from pyspark.sql import Window
from pyspark.sql.types import *


class Vintage():
    """
    计算 Vintage
    """
    def __init__(
            self, df, loan_id='id', loan_period_name='period',
            observe_date='20150228', observe_period=24, spark=None):
        """

        """
        self.df = df
        self.spark = spark
        self.loan_id = loan_id
        self.loan_period_name = loan_period_name
        self.observe_date = observe_date
        self.observe_period = observe_period
        self.df = self.df.withColumn(
            'loan_month',
            F.substring(F.col('loan_date'), 1, 6))
        self.df_month_total_loan = self.df.groupBy('loan_month').agg(
            F.sum(F.col('should_capital')).alias('should_capital'))

    def getObserveMonth(self, ):
        return pd.date_range(
            self.observe_date, periods=self.observe_period, freq='M'
        ).astype(str).str.replace('-', '').tolist()

    def calCapitalInterest(self, df_obs, obs_date):
        """
        1. 将还款时间大于记录时间的记为未还款
            a. 具体将已归零的剩余本息更换为应还本息
        2. 修改还款时间为记录时间
        3. 再将未完全还清本息的还款时间改为记录时间
        """
        df_obs = df_obs.withColumn(
            'surplus_capital',
            F.when(F.col('pay_date') > obs_date, F.col('should_capital')).otherwise(F.col('surplus_capital')))
        df_obs = df_obs.withColumn(
            'surplus_interest',
            F.when(F.col('pay_date') > obs_date, F.col('should_interest')).otherwise(F.col('surplus_interest')))
        return df_obs

    def calOverDue(self, df_obs, obs_date):
        """
        """
        # 计算逾期
        df_obs = df_obs.withColumn(
            'pay_date',
            F.when((F.col('pay_date') > obs_date) | (F.col('surplus_interest') > 0), obs_date).otherwise(
                F.col('pay_date')))
        df_obs = df_obs.withColumn(
            'due_days',
            F.datediff(F.to_timestamp(F.col('pay_date'), 'yyyyMMdd'), F.to_timestamp(F.col('due_date'), 'yyyyMMdd')))
        df_obs = df_obs.withColumn('due_days', F.when(F.col('due_days') < 0, 0).otherwise(F.col('due_days')))

        # 计算剩余本金
        window = Window.partitionBy(self.loan_id).orderBy(F.monotonically_increasing_id())
        df_obs = df_obs.withColumn(
            'paid_capital',
            F.sum(F.col('should_capital') - F.col('surplus_capital')).over(window))
        df_obs = df_obs.withColumn(
            'total_surplus_capital',
            F.col('total_loan') - F.col('paid_capital'))
        return df_obs

    def MOB(self, df_obs):
        @F.udf(returnType=StringType())
        def mob(x):
            if x == 0:
                return 'M0'
            elif x <= 30:
                return 'M1'
            elif x <= 60:
                return 'M2'
            elif x <= 90:
                return 'M3'
            else:
                return 'M4+'

        df_obs = df_obs.withColumn('mob', mob(F.col('due_days')))
        return df_obs

    def oneObserveDate(self, obs_date):
        """
        """
        df_obs = self.df.where(F.col('due_date') <= obs_date)
        df_obs = df_obs.fillna(str(obs_date), subset='pay_date')
        df_obs = self.calCapitalInterest(df_obs, obs_date)
        df_obs = self.calOverDue(df_obs, obs_date)
        df_obs = self.MOB(df_obs)
        # 剩余利息计算至 ID 级
        df_obs = df_obs.withColumn(
            'pay_int',
            F.sum(F.col('surplus_interest')).over(Window.partitionBy(self.loan_id)))
        df_not_pay_int = df_obs.where((F.col('pay_int') > 0) & (F.col('surplus_interest') > 0))
        df_paid_int = df_obs.where(F.col('pay_int') == 0)

        # 1. 有未结清利息，取最早一条
        # 2. 无未结清利息，取最新一条
        df_not_pay_int = df_not_pay_int.withColumn(
            'period_rank',
            F.row_number().over(Window.partitionBy(self.loan_id).orderBy(F.col(self.loan_period_name))))
        df_paid_int = df_paid_int.withColumn(
            'period_rank',
            F.row_number().over(Window.partitionBy(self.loan_id).orderBy(F.col(self.loan_period_name).desc())))

        df_obs = df_not_pay_int.where(F.col('period_rank') == 1).unionAll(
            df_paid_int.where(F.col('period_rank') == 1))

        # 计算坏账
        df_obs = df_obs.withColumn(
            'bad_amt',
            F.when(F.col('mob') == 'M4+', F.col('total_surplus_capital')).otherwise(0))
        df_obs_bad_amt_by_month = df_obs.groupBy('loan_month').agg(
            F.sum('bad_amt').alias('bad_amt'))

        # 账期
        df_obs_bad_amt_by_month = df_obs_bad_amt_by_month.withColumn(
            'book_month',
            F.months_between(
                F.to_date(F.lit(obs_date), 'yyyyMMdd'),
                F.to_date(F.col('loan_month'), 'yyyyMM')).cast(IntegerType()))
        df_obs_bad_amt_by_month = df_obs_bad_amt_by_month.withColumn(
            'observe_date', F.lit(obs_date))
        return df_obs_bad_amt_by_month

    def vintage(self, ):
        """
        """
        df_all_obs = list(map(self.oneObserveDate, self.getObserveMonth()))
        M4Plus = None
        for df_one_obs in df_all_obs:
            if M4Plus is None:
                M4Plus = df_one_obs
            else:
                M4Plus = M4Plus.unionAll(df_one_obs)
        return M4Plus

class VintagePd():
    """
    计算 Vintage
    """
    def __init__(
            self, df, loan_id='id', loan_period_name='period', observe_date='20150228',
            observe_period=24, loan_date_name='loan_date', mob='M4+',
            should_capital_name='should_capital'):
        '''
        描述：计算 Vintage 初始化

        参数：
        :param df: 借据表
        :param loan_id: 借据唯一标识
        :param loan_period_name: 期数字段名称
        :param observe_date: 初始观察日期
        :param observe_period: 观察周期

        示例：

        '''
        self.df = df
        self.loan_id = loan_id
        self.loan_period_name = loan_period_name
        self.observe_date = observe_date
        self.observe_period = observe_period
        self.should_capital_name = should_capital_name
        self.mob = mob
        self.df['loan_month'] = self.df[loan_date_name].dt.strftime('%Y%m')
        self.df_month_total_loan = self.df[['loan_month', should_capital_name]].groupby(
            'loan_month').sum()
        self.df_month_total_loan.reset_index(inplace=True)

    def getObserveMonth(self, ):
        return pd.date_range(
            self.observe_date, periods=self.observe_period, freq='M'
        )

    def convert(self, df, obs_date):
        """

        """
        df.fillna({'pay_date': obs_date}, inplace=True)
        df.loc[df['pay_date'] > obs_date, 'surplus_capital'] = df.loc[df['pay_date'] > obs_date, 'should_capital']
        df.loc[df['pay_date'] > obs_date, 'surplus_interest'] = df.loc[df['pay_date'] > obs_date, 'should_interest']
        df['pay_date'] = df['pay_date'].apply(lambda x: obs_date if x > obs_date else x)
        df.loc[df['surplus_interest'] > 0, 'pay_date'] = obs_date

        # 计算逾期天数，并将提前还款的逾期天数修改为0。
        df['due_days'] = (df['pay_date'] - df['due_date']) / np.timedelta64(1, 'D')
        df.loc[df['due_days'] < 0, 'due_days'] = 0

        df['paid_capital'] = df['should_capital'] - df['surplus_capital']
        df['paid_capital'] = df.groupby(self.loan_id)['paid_capital'].transform('cumsum')
        df['total_surplus_capital'] = df['total_loan'] - df['paid_capital']
        return df

    def MOB(self, due_days):
        """

        """
        if due_days == 0:
            return 'M0'
        elif due_days <= 30:
            return 'M1'
        elif due_days <= 60:
            return 'M2'
        elif due_days <= 90:
            return 'M3'
        else:
            return 'M4+'

    def oneObserveDate(self, obs_date):
        """

        """
        df_obs = self.df[self.df['due_date'] <= obs_date].copy()
        df_obs = self.convert(df_obs, obs_date)
        df_obs['mob'] = df_obs['due_days'].apply(self.MOB)

        df_obs['not_pay_int'] = df_obs.groupby(self.loan_id)['surplus_interest'].transform('sum')
        df_obs['period_min'] = df_obs.groupby(self.loan_id)['period'].transform('min')
        df_obs['period_max'] = df_obs.groupby(self.loan_id)['period'].transform('max')

        condition_paid_int = (df_obs.not_pay_int == 0) & (df_obs.period_max == df_obs.period)
        condition_not_pay_int = (df_obs.not_pay_int > 0) & (df_obs.period_min == df_obs.period) & (df_obs.surplus_interest > 0)
        df_obs = df_obs[condition_paid_int | condition_not_pay_int]

        df_obs['bad_amt'] = (df_obs['mob'] == self.mob) * df_obs['total_surplus_capital']

        df_obs_mob = df_obs.groupby('loan_month')[['bad_amt']].sum()
        df_obs_mob.reset_index(inplace=True)

        loan_date = pd.to_datetime(df_obs_mob.loan_month.astype(str) + pd.Series(['01'] * len(df_obs_mob)))
        df_obs_mob['book_month'] = ((obs_date - loan_date) / np.timedelta64(1, 'M')).round().astype(int)
        df_obs_mob['observe_date'] = obs_date
        return df_obs_mob

    def vintage(self, ):
        """
        """
        df_all_obs = list(map(self.oneObserveDate, self.getObserveMonth()))
        self.M4Plus = pd.DataFrame(columns=df_all_obs[0].columns)
        for df_one_obs in df_all_obs:
            self.M4Plus = self.M4Plus.append(df_one_obs)
        self.M4Plus = self.M4Plus.merge(self.df_month_total_loan, on='loan_month')
        self.M4Plus['bad_rate'] = self.M4Plus.bad_amt / self.M4Plus[self.should_capital_name]
        return self.M4Plus

    def vintagePivotTable(self, value):
        """
        :param value: `rate`、`amt`
        """
        if not hasattr(self, 'M4Plus'):
            self.vintage()
        return pd.pivot_table(
            self.M4Plus,
            values=f'bad_{value}',
            index='loan_month',
            columns='book_month',
            aggfunc='sum')
    #
    # def plot(self, ):
    #
    #     for col in LP_12_M4s.columns:
    #         LP_12_M4s[col] = (LP_12_M4s[col] / loan_LP.groupby('loan_month')['应还本金'].sum()).replace(0, np.nan)
    #     LP_12_M4s.iloc[: 12, :].transpose().plot(figsize=(12, 8), title='M4+ vintage')



from pyspark.sql import Column as TypeSparkColumn
import pandas as pd
import numpy as np
from pyspark.sql import functions as F
from pyspark.sql.types import StringType
from typing import List


__all__ = [
    "date_range",
    "month_shift",
    "to_date",
    "data_dates",
]


def date_range(start_date: str, end_date: str, freq: str = 'M') -> List[str]:
    """
    描述：时间周期

    :param start_date: 开始时间
    :param end_date: 结束时间
    :return:

    >>> dates = date_range('20191231', '20201231')
    """
    dates = pd.date_range(
        start=start_date, end=end_date, freq=freq
    ).astype(str).str.replace('-', '').tolist()
    return dates


def month_shift(col: str, n_month: int, fmt: str='yyyyMMdd') -> TypeSparkColumn:
    """
    描述：月份平移
    :param col:
    :param n_month:
    :param fmt:
    :return:
    """
    return F.date_format(F.add_months(F.to_date(col, fmt), n_month), fmt).cast(StringType())


def to_date(df, column, fmt: str='yyyyMMdd'):
    """"""
    df = df.withColumn(column, F.to_date(column, fmt))
    return df


def data_dates(obs_months=12, start_date='20191231', end_date='20210731', start_data_date='20190101'):
    """
    描述：计算样本批次，用于 UUID Info 长表与宽表的构建，用于样本的批量提取。

    参数：
    :param obs_months[Int]:          观察期（表现期）月份数
    :param start_date[String]:       建模样本开始时间，首批样本时间
    :param end_date[String]:         样本表现期最终批次时间点
    :param start_data_date[String]:  数据初始日期

    示例：
    >>> _data_dates, _sample_dates = data_dates(
    >>>     obs_months=2, start_date='20191031',
    >>>     end_date='20210731', start_data_date='20190101')
    """
    months = pd.date_range(start_data_date, end_date, freq='M').astype(str).str.replace('-', '').tolist()
    obs_date_index = months.index(start_date)
    credit_date_index = obs_date_index + obs_months

    data_dates = list(zip(months[: -(obs_months - 1)], months[(obs_months - 1):]))

    sample_dates = []
    while credit_date_index < len(months):
        sample_dates.append([months[obs_date_index], months[credit_date_index]])
        obs_date_index += 1
        credit_date_index += 1

    #### 剔除观察期不满足的记录
    data_start_date = np.array(sample_dates)[:, 0]
    mask = data_start_date >= months[obs_months - 1]
    sample_dates = np.array(sample_dates)[mask].tolist()
    return data_dates, sample_dates





__all__ = [
    'configs',
]


configs = [
    {
        'index_name': 'age',                 # 指标英文名
        'index_range': [[0, 30], [25, 80]],  # 指标值域，坏客户值域，好客户值域
        'index_type': 'randint',             # 数值类型
        'na_rate': [0.1, 0.2],
    },
    {
        'index_name': 'income',              # 指标英文名
        'index_range': [[500, 10000], [5000, 1e5]],  # 指标值域，坏客户值域，好客户值域
        'index_type': 'uniform',             # 数值类型
        'na_rate': [0.1, 0.2],
    },
]






import warnings
warnings.filterwarnings('ignore')


import numpy as np
import statsmodels.stats.api as sms
from tqdm import tqdm
import statsmodels.api as sm
from pandas import DataFrame, Series
from statsmodels.compat import lzip
from itertools import combinations, product
from statsmodels.stats.outliers_influence import OLSInfluence
from collections import defaultdict
import matplotlib.pyplot as plt


class LinearRegression:

    def __init__(self, data: DataFrame, y_name='endog', estimate='ols', log_cols=None):
        """
        Des: 线性回归方法
        :param data: [DataFrame] 数据
        :param y_name: [str] y 在数据集DataFrame中的名称
        :param log_cols: [str, list] 对给定的列取对数，可以是list给定列名称，
            也可以是字符串:
                'all', 全部取对数
                'endog', 因变量(y)取对数
                'exdog', 自变量(X)取对数
        :param estimate: [str] 估计方式 'ols' 普通最小二乘 or 'rlm' Robust lm
        """
        if isinstance(log_cols, list):
            self._log(log_cols, data)
            if y_name in log_cols:
                y_name = 'log_' + y_name

        elif log_cols == 'exdog':
            exdog_cols = data.columns.to_list()
            exdog_cols.remove(y_name)
            self._log(exdog_cols, data)

        elif log_cols == 'endog':
            data[y_name] = np.log(data[y_name])
            data.rename(columns={y_name: 'log_' + y_name}, inplace=True)
            y_name = 'log_' + y_name

        elif log_cols == 'all':
            cols = data.columns.to_list()
            self._log(cols, data)
            y_name = 'log_' + y_name
        else:
            Exception(ValueError, 'log_col value error!')


        y = data.pop(y_name)
        self.exog = data.copy()
        self.endog =y.copy()
        # 估计方式OLS or RLM
        self.estimate = estimate


    def _log(self, cols, data):
        """
        Des: 对列取对数
        :param cols:
        :param data:
        :return: None
        """
        for col in cols:
            data['log_' + col] = np.log(data[col])
        data.drop(cols, axis=1, inplace=True)


    def ols(self, X, y, constant=True):
        """
        Des: 最小二乘估计
        :param X: 数据X
        :param y: 数据y
        :param constant: 是否添加常数项
        :return: 回归结果
        """
        if constant and ('const' not in X.columns):
            X = sm.add_constant(X)
        return sm.OLS(y, X).fit()


    def rlm(self, X, y, constant=True):
        """
        Des: 最小二乘估计
        :param X: 数据X
        :param y: 数据y
        :param constant: 是否添加常数项
        :return: 回归结果
        """
        if constant and ('const' not in X.columns):
            X = sm.add_constant(X)
        return sm.RLM(y, X, M=sm.robust.norms.HuberT()).fit()


    def metrics(self, res):
        if self.estimate == 'ols':
            if self.exog.shape[-1] != 1:
                self.cond_num = self.condition_number(self.exog)
                self.vif_value = self.vif(self.exog)
            self.heteroske_test = self.heteroskedasticity(res)
            self.residuals_norm_test = self.residuals_norm(res)
            self.linearity_test = self.linearity(res)
            self.influence_test = self.influence(res)
            self.DW = self.dw(res)
        if self.estimate == 'rlm':
            raise Exception(ValueError, 'rlm 的回归诊断还未开发！')


    def vif(self, exog, exog_idx=None, ls=False):
        """
        Des: 方差膨胀因子 - 多重共线性参考指标
        :param exog: 变量
        :param exog_idx: 变量索引，在不指定变量索引时计算全部变量的vif，并以字典形式返回计算结果
        :param ls: 以list的形式返回数据
        :return: float or dict vif
        """
        from statsmodels.stats.outliers_influence import variance_inflation_factor
        if exog_idx:
            return variance_inflation_factor(exog, exog_idx)
        elif ls:
            return [variance_inflation_factor(exog.values, i) for i in range(exog.shape[-1])]
        else:
            vif_dict = {}
            for i, name in enumerate(exog):
                if name == "const":
                    continue
                vif_dict[name] = variance_inflation_factor(exog.values, i)
            return vif_dict


    def condition_number(self, exog):
        """
        Des: 条件数 - 多重共线性参考指标
        :param exog: 变量
        :return: [float] 条件数
        """
        return np.linalg.cond(exog)


    def residuals_norm(self, res):
        """
        Des: 残差正态性检验，Jarque-Bera，Omni-score 两种方法
        :param res: 回归结果
        :return: [dict]
        """
        residuals_norm_test_res = {}

        name = ['Jarque-Bera', 'Chi^2 two-tail prob.', 'Skew', 'Kurtosis']
        test = sms.jarque_bera(res.resid)
        residuals_norm_test_res['Jarque-Bera'] = lzip(name, test)

        name = ['Chi^2 score', 'Two-tail probability']
        test = sms.omni_normtest(res.resid)
        residuals_norm_test_res['Omni'] = lzip(name, test)
        return residuals_norm_test_res


    def heteroskedasticity(self, res):
        """
        Des: 异方差检验 Breush - Pagan 与 Goldfeld - Quandt两种检验方法
        :param res: 回归结果
        :return: [dict]
        """
        # Breush - Pagan
        heteroske_test_res = {}
        name = ['Lagrange multiplier statistic', 'p-value',
                'f-value', 'f p-value']
        test = sms.het_breuschpagan(res.resid, res.model.exog)
        heteroske_test_res['Breush-Pagan'] = lzip(name, test)

        # Goldfeld - Quandt
        name = ['F statistic', 'p-value']
        test = sms.het_goldfeldquandt(res.resid, res.model.exog)
        heteroske_test_res['Goldfeld-Quandt'] = lzip(name, test)
        return heteroske_test_res


    def linearity(self, res):
        """
        Des: 线性性检验
        :param res: 回归结果
        :return: [list]
        """
        name = ['t value', 'p value']
        test = sms.linear_harvey_collier(res)
        return lzip(name, test)


    def influence(self, res):
        """
        Des: 影响测试
        :param res: 回归结果
        :return: DataFrame
        """
        influence_test = OLSInfluence(res)
        return influence_test.summary_frame()


    # Plot leverage statistics vs. normalized residuals squared
    def plot_leverage_resid2(self, res):
        """
        Des: leverage vs. resid
        :param res: 回归结果
        :return: plt
        """
        sm.graphics.plot_leverage_resid2(res)
        return None


    def plot_variance_homogeneity(self, exog, resid, **kwargs):
        """

        :param exog:
        :param res:
        :param kwargs:
        :return:
        """
        if isinstance(exog, DataFrame):
            cols = exog.columns.to_list()
            kwargs.setdefault('figsize', (10, len(cols) * 3.5))

        kwargs.setdefault('style', 'seaborn')
        kwargs.setdefault('dpi', 150)
        kwargs.setdefault('labelfontsize', 10)
        kwargs.setdefault('tickfontsize', 8)
        kwargs.setdefault('titlefontsize', 12)
        kwargs.setdefault('x_label', 'False Positive Rate')
        kwargs.setdefault('y_label', 'True Positive Rate')
        kwargs.setdefault('title', 'Receiver operating characteristic')

        plt.style.use(kwargs.get('style'))

        if isinstance(exog, Series):
            plt.figure(figsize=(6, 4), dpi=kwargs.get('dpi'))
            plt.scatter(exog, resid)
            plt.axhline(y=0, color='tab:red', linestyle='--')
            plt.xlabel(exog.name)
            plt.ylabel('Residual')
            return None

        fig, axs = plt.subplots(
            len(cols),
            figsize=kwargs.get('figsize'),
            dpi=kwargs.get('dpi')
        )
        for i, col in enumerate(cols):
            axs[i].scatter(exog[col], resid)
            axs[i].set_xlabel(col)
            axs[i].axhline(y=0, color='tab:red', linestyle='--')
            axs[i].set_ylabel('Residual')
        return None


    def plot_resid(self):
        import seaborn as sns
        from scipy.stats import norm
        sns.distplot(
            self.res.resid,
            fit=norm,
            kde=False,
        )
        return None


    def plot_pp(self):
        pq = sm.ProbPlot(self.res.resid)
        pq.ppplot(line='45')
        return None


    def plot_qq(self):
        pq = sm.ProbPlot(self.res.resid)
        pq.qqplot(line='q')
        return None


    def dw(self, res):
        return sm.stats.stattools.durbin_watson(res.resid)


    def bp_test(self, ):
        return sm.stats.diagnostic.het_breuschpagan(self.res.resid, self.res.model.exog)


    def plot_regression(self, x, res, **kwargs):

        if res.params[0] <= 0:
            kwargs.setdefault('title', f'$y = {res.params[-1]: .2f} x {res.params[0]: .2f}$')
        else:
            kwargs.setdefault('title', f'$y = {res.params[-1]: .2f} x + {res.params[0]: .2f}$')

        kwargs.setdefault('xlabel', 'exdog')
        kwargs.setdefault('ylabel', 'endog')
        kwargs.setdefault('dpi', 100)
        kwargs.setdefault('figsize', (8, 6))

        fig, ax = plt.subplots(figsize=kwargs.get('figsize'), dpi=kwargs.get('dpi'))
        ax.plot(x, self.endog, 'o', label="data")
        ax.plot(x, res.fittedvalues, 'r--.', label="Regression")
        ax.set_xlabel(kwargs.get('xlabel'))
        ax.set_ylabel(kwargs.get('ylabel'))
        ax.set_title(kwargs.get('title'))
        ax.legend()


    def different_variable_combinations(self, *args):
        """
        Des: OLS Different variable combinations
        :param features_num: 变量数
        :return: model_index, variables_index
        """
        res_dict = defaultdict(list)
        cols = self.exog.columns.to_list()

        variables_index = DataFrame()

        if "const" in cols:
            cols.remove("const")

        combins = None

        if isinstance(args[0], int):
            features_num = args[0]
            if len(cols) < features_num:
                raise Exception(ValueError, 'features_num must more than the columns!')
            combins = list(enumerate(combinations(cols, features_num)))
        elif args:
            features_num=None
            combins = list(enumerate(product(*args)))

        if combins:
            for i, temp_cols in tqdm(combins, desc=f'Total regressions {len(combins)}'):
                exog = self.exog[list(temp_cols)]
                res = self.ols(exog, self.endog)
                res_dict['columns'].append(temp_cols)
                res_dict['nobs'].append(res.nobs)
                res_dict['rsquared'].append(res.rsquared)
                res_dict['rsquared_adj'].append(res.rsquared_adj)
                res_dict['f_pvalue'].append(res.f_pvalue)
                res_dict['aic'].append(res.aic)
                res_dict['bic'].append(res.bic)
                res_dict['DW'].append(self.dw(res))

                if features_num != 1:
                    variables_index[f'group_{i}'] = res.pvalues.index
                    variables_index[f'pvalues_{i}'] = res.pvalues.values
                    vifLs = self.vif(exog, ls=True)
                    vifLs.insert(0, np.nan)
                    variables_index[f'vif_{i}'] = vifLs

            return DataFrame(res_dict), variables_index
        else:
            raise Exception(ValueError, 'combins are None!')




import numpy as np
import pandas as pd

def PDToRant(pd: float, ranking: list, max_grade: int = 16):
    """
    PD 到主标尺等级的映射
    :param pd: 预测PD
    :param ranking: 主标尺分箱的max bin
    :return: 等级
    """
    if pd==1.:
        return max_grade
    else:
        return sum(np.array(ranking) <= pd) + 1



import numpy as np
import pandas as pd
from pandas import Series, DataFrame
from pandas.core.groupby.generic import DataFrameGroupBy
from typing import Any, List, Union, Optional, Sequence
from ..utils.logger import LoggerMixin
from .utils import *


__all__ = [
    "WOEMixin",
]


class WOEMixin(LoggerMixin):
    """
    woe Mixin
    todo: 增加 bins plot
    """
    df_data: DataFrame
    df_missing_data: DataFrame
    df_abnormal_data: DataFrame
    woe: DataFrame
    woe_normal: DataFrame
    woe_missing: DataFrame
    woe_abnormal: DataFrame
    woe_columns: Optional[List[str]]
    decimal: Optional[int] = 6

    def _init__(
            self,
            variables: Series,
            target: Series,
            var_name: Optional[str] = None,
            fill_bin: Optional[bool] = True,
            abnormal_vals: Optional[List[Union[str, int, float]]] = None,
            weight: Optional[Any] = None,
            woe_columns: Optional[List[str]] = None,
            *args: Any,
            **kwargs: Any,
    ):
        """
        初始化
        :param variables: 待分箱变量
        :param target: 标签
        :param bins: 分箱数量
        :param var_name: 变量名称
        :param fill_bin: 是否在分箱中未包含 1 标签的情况下，全部 1 标签数量自增 1 默认为 False
        :param abnormal_vals: 特殊值分箱，在变量存在特殊值时单独分一箱，如 -1111, -9999。
        :param weight: 样本权重。
        """
        self.variables = variables
        self.target = target
        self.fill_bin = fill_bin
        self.abnormal_vals = abnormal_vals
        self.woe_columns = woe_columns
        self.args = args
        self.kwargs = kwargs

        self._get_var_name(var_name)
        self._validate_data_length(variables, target)

        self._basic_count(target, weight)
        self._data_prepare(variables, target, weight)

    def __call__(self, *args, **kwargs) -> DataFrame:
        """"""
        return self.fit()

    def fit(self) -> DataFrame:
        """"""
        return DataFrame()

    def _get_var_name(self, var_name: Optional[str] = None) -> None:
        """"""
        if hasattr(self.variables, 'name'):
            self.var_name = self.variables.name
        else:
            self.var_name = var_name

    def _validate_data_length(self, *args) -> None:
        """
        检查数据长度是否一致
        :param args: 待检查数据
        """
        length = [len(item) for item in args]
        if len(set(length)) == 1:
            pass
        else:
            raise ValueError('Variable and target must be the same length!')

    def _data_prepare(
            self,
            variables: Series,
            target: Series,
            weight: Optional[Any] = None
    ) -> None:
        """
        prepare data: 1. missing data; 2. abnormal data; 3. normal data
        """
        if weight is None:
            df = pd.DataFrame({"variables": variables, "target": target})
        else:
            df = pd.DataFrame({"variables": variables, "target": target, 'weight': weight})

        # drop infinite values
        if pd.api.types.is_number(df.variables):
            df = df[~np.isinf(df.variables)]

        if self.abnormal_vals:
            condition_abnormal = df.variables.isin(self.abnormal_vals)
            self.df_abnormal_data = df[condition_abnormal]
            self.df_missing_data = df[df.variables.isna() & ~condition_abnormal]
            self.df_data = df[df.variables.notna() & ~condition_abnormal]
        else:
            self.df_abnormal_data = pd.DataFrame()
            self.df_missing_data = df[df.variables.isna()]
            self.df_data = df[df.variables.notna()]

        if weight is not None:
            self.weight = self.df_data.weight
        else:
            self.weight = None

    def data_prepare(
            self,
            variables: Optional[Series],
            target: Optional[Series],
            weight: Optional[Series] = None,
            var_name: Optional[str] = None,
    ):
        """"""
        self._get_var_name(var_name)
        self._validate_data_length(variables, target)
        self._basic_count(target, weight)
        self._data_prepare(variables, target, weight)
        self.woe_columns = [
            'var_name', 'missing_rate', 'min_bin', 'max_bin', 'total',
            'bad', 'bad_rate', 'woe', 'iv', 'iv_value',
        ]

    def _basic_count(self, target: Optional[Series], weight: Optional[Series] = None):
        """ basic count """
        self.sample_length = len(target)

        if weight is None:
            self.total, self.bad = target.count(), target.sum()
            self.good = self.total - self.bad
        else:
            self.total = weight.sum()
            self.bad = weight[target == 1].sum()
            self.good = weight[target == 0].sum()

    def _bucket_stats(self, bucket: Optional[Series] = None) -> None:
        """ 基本结果表 """
        if len(bucket) > 0:
            self.woe_normal = pd.DataFrame({
                'min_bin': bucket.variables.min(),
                'max_bin': bucket.variables.max(),
                'bad': bucket.target.sum(),
                'good': bucket.target.count() - bucket.target.sum(),
                'total': bucket.target.count(),
            })
        else:
            self.woe_normal = pd.DataFrame()

        if len(self.df_missing_data) != 0:
            self.woe_missing = pd.DataFrame({
                'min_bin': [np.nan],
                'max_bin': [np.nan],
                "bad": self.df_missing_data.target.sum(),
                'good': self.df_missing_data.target.count() - self.df_missing_data.target.sum(),
                'total': self.df_missing_data.target.count(),
            })
        else:
            self.woe_missing = pd.DataFrame()

        if len(self.df_abnormal_data) != 0:
            self.woe_abnormal = self.df_abnormal_data.groupby(
                'variables').agg({'target': ['sum', 'count']})
            self.woe_abnormal.columns = ['bad', 'total']
            self.woe_abnormal['min_bin'] = self.woe_abnormal.index
            self.woe_abnormal['max_bin'] = self.woe_abnormal.index
            self.woe_abnormal['good'] = self.woe_abnormal.total - self.woe_abnormal.bad
            self.woe_abnormal = self.woe_abnormal[[
                'min_bin', 'max_bin', "bad", 'good', 'total',]]
            self.woe_abnormal.reset_index(inplace=True, drop=True)
        else:
            self.woe_abnormal = pd.DataFrame()

    def get_weighted_buckets(
            self,
            value_cut: Optional[Sequence] = None,
            df: Optional[DataFrame] = None,
            variables: Optional[Series] = None,
            target: Optional[Series] = None,
            weight: Optional[Series] = None,
    ):
        """获取加权分组"""
        if df is not None:
            variables = df.variables
            target = df.target
            weight = df.weight
        else:
            pass

        weight_good, weight_bad = weight.copy(), weight.copy()
        weight_good[target == 1], weight_bad[target == 0] = 0, 0

        if value_cut is not None:
            bucket = pd.DataFrame({
                'variables': variables,
                'target': target,
                'weight': weight,
                'weight_good': weight_good,
                'weight_bad': weight_bad,
                'bucket': value_cut,
            }).groupby('bucket', as_index=True, observed=False)
        else:
            bucket = pd.DataFrame({
                'variables': variables,
                'target': target,
                'weight': weight,
                'weight_good': weight_good,
                'weight_bad': weight_bad,
            }).groupby('variables', as_index=True, observed=False)
        return bucket

    def _weighted_buckets_stats(self, bucket):
        """weighted good or bad"""
        if len(bucket) > 0:
            self.woe_normal = pd.DataFrame({
                'min_bin': bucket.variables.min(),
                'max_bin': bucket.variables.max(),
                'bad': bucket.weight_bad.sum(),
                'good': bucket.weight_good.sum(),
                'total': bucket.weight.sum(),
            })
        else:
            self.woe_normal = pd.DataFrame()

        if len(self.df_missing_data) != 0:
            self.woe_missing = pd.DataFrame({
                'min_bin': [np.nan],
                'max_bin': [np.nan],
                "bad": self.df_missing_data.weight[self.df_missing_data.target == 1].sum(),
                'good': self.df_missing_data.weight[self.df_missing_data.target == 0].sum(),
                'total': self.df_missing_data.weight.sum(),
            })
        else:
            self.woe_missing = pd.DataFrame()

        if len(self.df_abnormal_data) != 0:
            abnormal_bucket = self.get_weighted_buckets(df=self.df_abnormal_data)
            self.woe_abnormal = pd.DataFrame({
                'min_bin': abnormal_bucket.target.min().index,
                'max_bin': abnormal_bucket.target.max().index,
                'bad': abnormal_bucket.weight_bad.sum(),
                'good': abnormal_bucket.weight_good.sum(),
                'total': abnormal_bucket.weight.sum(),
            })
        else:
            self.woe_abnormal = pd.DataFrame()

    def calculate_woe(
            self,
            woe: DataFrame,
            bucket_type: Optional[int] = 0,
            reset_boundary: Optional[bool] = False
    ) -> DataFrame:
        """
        依据结果表，计算woe
            如果存在 fill_bin==True, 则需要对分箱中未包含 1 标签的情况下，全部 1 标签数量自增 0.5
        """
        if woe.empty:
            return woe
        else:
            if self.fill_bin and (0 in woe.bad.tolist()):
                woe.bad = woe.bad + 0.5
            elif self.fill_bin and (0 in woe.good.tolist()):
                woe.good = woe.good + 0.5
            woe['bad_rate'] = woe['bad'] / woe['total']
            woe['bad_attr'] = woe['bad'] / self.bad
            woe['good_attr'] = woe['good'] / self.good
            woe['woe'] = np.log(woe['bad_attr'] / woe['good_attr'])
            woe['iv'] = (woe['bad_attr'] - woe['good_attr']) * woe['woe']
            woe = woe.sort_values(by='min_bin').reset_index(drop=True)

            if isinstance(bucket_type, int):
                woe["bucket_type"] = bucket_type
            if reset_boundary:
                woe = self._reset_woe_boundary(woe)
        return woe

    def _add_order(self, order: Optional[bool] = True) -> None:
        """"""
        if order and "order" not in self.woe_columns:
            self.woe_columns.append('order')
            self.woe['order'] = monotony(self.woe_normal.woe)

    def _add_score(self, score: Optional[bool] = True) -> None:
        """"""
        if score and "score" not in self.woe_columns:
            self.woe_columns.append('score')
            neg_woe = - self.woe.woe
            woe_score = (neg_woe - neg_woe.min()) * 100 / (neg_woe.max() - neg_woe.min())
            self.woe['score'] = list(map(int_score, woe_score))

    def _add_origin_border(self, origin_border: Optional[bool] = False) -> None:
        """"""
        if origin_border and "min_bin_val" not in self.woe_columns:
            self.woe_columns.extend(['min_bin_val', 'max_bin_val'])

    def _reset_woe_boundary(self, woe: DataFrame) -> DataFrame:
        """"""
        woe['min_bin_val'] = woe['min_bin']
        woe['max_bin_val'] = woe['max_bin']

        right = woe.max_bin.tolist()
        if isinstance(right[0], str):
            return woe
        elif isinstance(right[0], (float, int)):
            right[-1] = np.inf
            left = [-np.inf] + right[: -1]
            woe['min_bin'] = left
            woe['max_bin'] = right
            return woe
        else:
            raise ValueError("min_bin or max_bin is not str or float")

    def _merge_woe(self, *args: DataFrame) -> DataFrame:
        """"""
        woes = [woe for woe in args if not woe.empty]
        return pd.concat(woes, axis=0)

    def _value_decimal(self, woe: DataFrame) -> DataFrame:
        """"""
        decimal_columns = ["bad_rate", "woe", "iv", "iv_value"]
        for col in decimal_columns:
             woe[col] = woe[col].round(self.decimal)
        return woe

    def cal_woe_iv(
            self,
            bucket: Optional[Union[Series, DataFrameGroupBy]],
            score: Optional[bool] = True,
            origin_border: Optional[bool] = False,
            order: Optional[bool] = True
    ) -> None:
        """
        描述：计算 woe 结果。

        :param bucket:
        :param score: 是否增加 woe score。
        :param origin_border: 是否增加 分箱中的最大值与最小值。
        :param order: 是否按照 min_bin 排序。
        :return: DataFrame
        """
        if not hasattr(self, 'weight') or self.weight is None:
            self._bucket_stats(bucket=bucket)
        else:
            self._weighted_buckets_stats(bucket=bucket)

        self.woe_normal = self.calculate_woe(self.woe_normal, bucket_type=0, reset_boundary=True)
        self.woe_missing = self.calculate_woe(self.woe_missing, bucket_type=1)
        self.woe_abnormal = self.calculate_woe(self.woe_abnormal, bucket_type=2)

        self.woe = self._merge_woe(self.woe_normal, self.woe_abnormal, self.woe_missing)
        self.woe['iv_value'] = self.woe.iv[~np.isinf(self.woe.iv)].sum()
        self.woe['missing_rate'] = 1 - (self.woe_normal.total.sum() / self.sample_length)
        self.woe['var_name'] = self.var_name
        self.woe.reset_index(drop=True, inplace=True)

        self._add_order(order)
        self._add_score(score)
        self._add_origin_border(origin_border)
        self.woe = self._value_decimal(self.woe)
        self.woe = self.woe[self.woe_columns]



from typing import List, Dict, Union, Callable, Optional
try:
    from pyspark.sql import functions as F
    from pyspark.sql import SparkSession
    from pyspark.sql import DataFrame
    from pyspark.sql.types import FloatType
except ImportError:
    print("pyspark not installed, please install it first. < pip install pyspark >")
from itertools import product
from ..spark import *
from .schema import *
from .utils import *
from .date import *


__all__ = [
    "CollectIndex"
]


class CollectIndex(LoggerMixin):
    """"""
    index_schema_list = ['uuid', 'index_name', 'index_val', 'load_date']

    def __init__(
            self,
            df: Optional[DataFrame],
            index_fields: List[IndexField],
            spark: SparkSession,
            reset: Optional[bool] = False,
            group_mapping: Optional[Dict[str, str]] = None,
            operator_mapping: Optional[Dict[str, Callable]] = None,
            decimal: Optional[int] = 6,
            logger: Optional[Callable] = None,
            verbose: Optional[bool] = False,
    ):
        """
        :param df: DataFrame 原始数据
        :param index_fields: List[IndexField] 指标衍生配置清单
        :param spark: SparkSession
        :param reset: 是否重置指标表
        :param group_mapping: Dict[str, str] 分组字段映射 ent or other
        :param operator_mapping: Dict[str, Callable] 指标运算符映射
        :param decimal: 指标值保留小数位数
        :param logger: 日志
        :param verbose: 是否打印日志
        """
        self.df = df
        self.index_fields = index_fields
        self.spark = spark
        self.reset = reset
        self.group_mapping = group_mapping
        self.operator_mapping = operator_mapping
        self.decimal = decimal
        self.logger = logger
        self.verbose = verbose

    def __call__(self, *args, **kwargs):
        """"""
        self.run()

    def _get_data(self, df: DataFrame, column: Union[List[str], str], start_date: str, end_date: str):
        """"""
        self._logger(msg=f"[{__class__.__name__}] Get data from {column} between {start_date} and {end_date}]")
        df = df.where(F.col("load_date").between(start_date, end_date))
        self._logger(msg=f"[{__class__.__name__}] In fact load_date: {len(df.select('load_date').distinct().collect())}")
        select_columns = ["uuid", "load_date"]
        if isinstance(column, str):
            select_columns.append(column)
        else:
            select_columns.extend(column)
        return df.select(select_columns)

    # def _run_core(self, df_origin: DataFrame) -> DataFrame:
    #     """"""
    #     df_origin = get_operator_data(df=)
    #     return df_origin

    def _run_basic(self):
        """"""

    def _run_qoq(self):
        """"""

    def _run_yoy(self):
        """"""

    def _save_index(self, df: DataFrame, index_field: IndexField, load_date: str, index_name: str):
        """"""
        insert_into_hive(
            df, db_name=index_field.save_db, table_name=index_field.save_table, spark=self.spark,
            partition=['load_date', 'index_name'], partition_val=[load_date, index_name],
        )
        self._logger(msg=f"[{__class__.__name__}] Save index data to {index_field.save_db}.{index_field.save_table}")

    def _one_column(
            self,
            index_field: IndexField,
            operate_name: str,
            month: Union[int, str],
            id_level: str,
            alias_operator: Optional[str] = None
    ):
        """"""
        collect_index_name = merge_index_name(
            id_level=id_level, index_name=index_field.index_name,
            operate_name=operate_name, month=month, alias_operator=alias_operator)
        self._logger(msg=f"[{__class__.__name__}] current index name: {collect_index_name}\n")
        start_date = get_start_and_end_date(biz_date=index_field.load_date, month=month)
        self.df_origin = self._get_data(
            df=self.df, column=index_field.columns, start_date=start_date, end_date=index_field.load_date)
        group_col = self.group_mapping.get(id_level)

        if isinstance(index_field.columns, str):
            agg_columns = [index_field.columns]
        else:
            agg_columns = index_field.columns
        agg_columns = [F.col(column).cast(FloatType()) for column in agg_columns]
        df_agg = self.df_origin.groupBy(group_col).agg(
            self.operator_mapping.get(operate_name)(*agg_columns).alias("index_val")
        )
        df_agg = df_agg.withColumn("index_val", F.round(df_agg["index_val"], self.decimal))
        df_agg = df_agg.withColumn("index_name", F.lit(collect_index_name))
        df_agg = df_agg.withColumn("load_date", F.lit(index_field.load_date))
        df_agg = df_agg.select(self.index_schema_list)
        self._save_index(
            df=df_agg, index_field=index_field, load_date=index_field.load_date, index_name=collect_index_name)
        return df_agg

    def _loop_load_date(
            self,
            index_field: IndexField,
            operate_name: str,
            month: Union[int, str],
            id_level: str
    ) -> DataFrame:
        """"""
        if index_field.alias_operators:
            alias_operator = index_field.alias_operators[index_field.operators.index(operate_name)]
        else:
            alias_operator = None
        df_agg = self._one_column(
            index_field=index_field, operate_name=operate_name, month=month, id_level=id_level,
            alias_operator=alias_operator)
        return df_agg

    def _loop_date_periods(self):
        """"""

    def run(self):
        """"""
        for index_field in self.index_fields:
            for operate_name, month, id_level in product(index_field.operators, index_field.months, index_field.id_levels):
                if index_field.load_date:
                    df_agg = self._loop_load_date(index_field=index_field, operate_name=operate_name, month=month, id_level=id_level)
                elif index_field.start_date and index_field.end_date:
                    pass
                else:
                    raise Exception("[start_date/end_date] or [load_date] must be set.")



import numpy as np
import pandas as pd
import seaborn as sns
from typing import Any, List, Union, Sequence, Optional
import matplotlib.pyplot as plt
from scipy.stats import ks_2samp, t
from sklearn.feature_selection import chi2
from sklearn.metrics import (
    roc_auc_score, roc_curve, recall_score, precision_score, confusion_matrix, f1_score,
    precision_recall_curve
)
from .schema import *
from .utils import lift


__all__ = [
    'ks_test',
    "KSValue",
    'Chi2',
    'BinaryMetrics',
    'p_value',
    "BinaryMetrics",
]


def ks_test(feature, target):
    """ """
    try:
        return ks_2samp(feature[target == 1], feature[target == 0])
    except:
        return [-1, -1]


def KSValue(variables, target):
    """
    Des: 计算KS值，计算KS-test的P-value
    :param variables:
    :param target:
    :return: ks_value, p_value
    """
    from scipy.stats import ks_2samp
    get_ks = lambda variables, target: ks_2samp(variables[target == 1], variables[target == 0])
    ks_value, p_value = get_ks(variables, target)
    return ks_value, p_value


def Chi2(X, y):
    """ """
    try:
        chi_statistic, chi_pvalue = chi2(X, y)
        return float(chi_statistic[0]), float(chi_pvalue[0])
    except:
        return -1, -1


class BinaryMetrics:
    metrics_output: BinaryMetricsOutput

    def __init__(
            self,
            target: Sequence,
            proba: Optional[Sequence] = None,
            prediction: Optional[Sequence] = None,
            score: Optional[Sequence] = None,
            **kwargs: Any,
    ):
        """
        :param target: 真实标签
        :param proba: 预测概率
        :param prediction: 预测标签
        :param score: 预测得分
        :param kwargs: 其他参数
        """
        self.target = target
        self.proba = proba
        self.prediction = prediction
        self.score = score
        self.metrics_output = BinaryMetricsOutput()
        self.kwargs = kwargs
        self.metrics_method = []
        self._validate_data()

    def _validate_data(self):
        """"""
        value_length = [len(item) for item in [self.target, self.proba, self.prediction, self.score] if item is not None]
        if value_length[0] == 0 or len(set(value_length)) > 1:
            raise ValueError("All input data must have the same length")

        if self.prediction is None and self.proba is None:
            raise ValueError("Either prediction or proba must be provided")
        if self.proba is not None:
            self.metrics_method.extend([
                "auc", "roc_curve",
            ])
        if self.prediction is not None:
            self.metrics_method.extend([
                "recall", "f1_score", "precision", "pr_curve",
                "confusion_matrix",
            ])

    def auc(self):
        """"""
        self.metrics_output.auc = roc_auc_score(y_true=self.target, y_score=self.proba)

    def roc_curve(self):
        """"""
        fpr, tpr, thresholds = roc_curve(y_true=self.target, y_score=self.proba)
        _roc_curve = ROCCurveOutput(fpr=fpr.tolist(), tpr=tpr.tolist(), thresholds=thresholds)
        self.metrics_output.roc_curve = _roc_curve
        self.metrics_output.ks = float(abs(fpr - tpr).max())

    def pr_curve(self):
        """"""
        precision, recall, thresholds = precision_recall_curve(y_true=self.target, probas_pred=self.proba)
        _pr_curve = PrecisionRecallCurveOutput(
            precision=precision.tolist(), recall=recall.tolist(), thresholds=thresholds)
        self.metrics_output.pr_curve = _pr_curve

    def recall(self):
        """"""
        self.metrics_output.recall = recall_score(y_true=self.target, y_pred=self.prediction)

    def f1_score(self):
        """"""
        self.metrics_output.f1_score = f1_score(y_true=self.target, y_pred=self.prediction)

    def precision(self):
        """"""
        self.metrics_output.precision = precision_score(y_true=self.target, y_pred=self.prediction)
        self.metrics_output.lift = float(self.metrics_output.precision / np.array(self.target).mean())

    def confusion_matrix(self):
        """"""
        tn, fp, fn, tp = confusion_matrix(y_true=self.target, y_pred=self.prediction).ravel()
        self.metrics_output.confusion_matrix = ConfusionMatrixOutput(tn=tn, fp=fp, fn=fn, tp=tp)

    def evaluate(self) -> BinaryMetricsOutput:
        """"""
        for metrics in self.metrics_method:
            getattr(self, metrics)()
        return self.metrics_output

    def plot_roc(self):
        """ Des: 绘制roc图 """
        if self.metrics_output.roc_curve is None or self.metrics_output.auc is None:
            raise ValueError('auc/roc is None, please run evaluate first')

        max_idx = np.argmax(abs(np.array(self.metrics_output.roc_curve.tpr) - np.array(self.metrics_output.roc_curve.fpr)))

        plt.plot(
            self.metrics_output.roc_curve.fpr, self.metrics_output.roc_curve.tpr,
            color='darkorange', lw=2, label=f'ROC curve (area = {self.metrics_output.auc:.2f}, ks = {self.metrics_output.ks:.2f})')
        plt.scatter(
            self.metrics_output.roc_curve.fpr[max_idx],
            self.metrics_output.roc_curve.tpr[max_idx],
            alpha=0.7, marker='*', s=50, color='tab:red', label=r'$ks\ max\ point$'
        ).set_zorder(10)

        plt.plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--')
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.0])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('Receiver operating characteristic')
        plt.legend(loc="lower right")
        plt.show()

    def plot_pr(self):
        """
        Des: 绘制pr图
        :return: None
        """
        if self.metrics_output.pr_curve is None or self.metrics_output.auc is None:
            raise ValueError('auc/pr is None, please run evaluate first')
        plt.plot(
            self.metrics_output.pr_curve.recall,
            self.metrics_output.pr_curve.precision, color='darkorange',
            lw=2, label=f'PR curve (area = {self.metrics_output.auc:.2f})')
        plt.plot([0, 1], [1, 0], color='navy', lw=2, linestyle='--')
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.0])
        plt.xlabel('Recall')
        plt.ylabel('Precision')
        plt.title('Precision-Recall Curve')
        plt.legend(loc="lower right")
        plt.show()

    def plot_confusion_matrix(self, normalized=False):
        """
        Des: 绘制混淆矩阵图
        :param normalized: 是否归一化，默认为False
        """
        if self.metrics_output.confusion_matrix is None:
            raise ValueError("Please call evaluate() first")

        matrix = np.array([
            [self.metrics_output.confusion_matrix.tp, self.metrics_output.confusion_matrix.fp],
            [self.metrics_output.confusion_matrix.fn, self.metrics_output.confusion_matrix.tn],
        ])

        if normalized:
            matrix = (matrix - matrix.min()) / (matrix.max() - matrix.min())
        ax = sns.heatmap(matrix, annot=True, cmap="YlGnBu")
        ax.set_xlabel('True label')
        ax.set_xticklabels(["P", "N"])
        ax.set_ylabel('Predicted label')
        ax.set_yticklabels(["P", "N"])
        ax.set_title('Confusion matrix')
        plt.show()

    def plot_score(self, **kwargs):
        """"""
        """"""
        if self.score is None:
            raise ValueError("Score is None.")

        sns.set_style("whitegrid")
        fig, axs = plt.subplots()
        good_score = self.score[self.target == 0]
        bad_score = self.score[self.target == 1]
        with sns.color_palette("muted", n_colors=2):
            sns.histplot(good_score, label="good", kde=True, stat="percent", **kwargs)
            sns.histplot(bad_score, label="bad", kde=True, stat="percent", **kwargs)
        axs.legend()
        axs.set_title(r"$Score\ Distribution$")
        axs.set_xlabel('$Score$')
        axs.set_ylabel('$Percent$')
        plt.show()

    def plot_ks(
            self,
            bins: Optional[Union[int, List[Union[int, float]]]] = 40,
            show_bucket: Optional[bool] = False,
    ):
        """"""
        df_score = pd.DataFrame({"score": self.score, "target": self.target})
        df_lift = lift(df_score, bins=bins)
        x = range(len(df_lift.index.tolist()))
        mark_line_good = df_lift.loc[df_lift.ks == df_lift.ks.max(), "cum_good"].values[0]
        mark_line_bad = df_lift.loc[df_lift.ks == df_lift.ks.max(), "cum_bad"].values[0]

        fig, axs = plt.subplots()
        axs.plot(x, df_lift.ks * 100, marker='o', ms=2, linewidth=1.5, c='tab:blue', label="ks")
        axs.plot(x, df_lift.cum_good * 100, marker='o', ms=2, linewidth=1.5, c='tab:green', label="cum_good")
        axs.plot(x, df_lift.cum_bad * 100, marker='o', ms=2, linewidth=1.5, c='tab:red', label="cum_bad")

        axs.axvline(x=np.argmax(df_lift.ks.to_list()), linestyle='--', linewidth=0.7, alpha=0.5, c='gray')
        axs.axhline(y=df_lift.ks.max() * 100, linestyle='--', linewidth=0.7, alpha=0.5, c='tab:blue')
        axs.axhline(y=mark_line_good * 100, linestyle='--', linewidth=0.7, alpha=0.5, c='tab:green')
        axs.axhline(y=mark_line_bad * 100, linestyle='--', linewidth=0.7, alpha=0.5, c='tab:red')

        axs.set_title(f'KS = {df_lift.ks.max(): .2f} at {bins} bucket')
        axs.legend()
        axs.set_xlabel("buckets")
        axs.set_ylabel(r'$Rate(\%)$')
        if show_bucket:
            axs.set_xticks(range(len(df_lift.index)))
            axs.set_xticklabels(df_lift.index.tolist(), rotation=90)
        plt.tight_layout()
        plt.show()


def p_value(params, x, y, predict):
    """ 描述：计算广义线性回归 std_error t_value p_vale """
    x = np.hstack([np.ones(len(x), 1), x])
    n_degree = (y.shape[0] - x.shape[1])

    mse = sum((y - predict) ** 2) / n_degree
    var_error = mse * np.linalg.inv(np.dot(x.T, x)).diagonal()
    std_error = np.sqrt(var_error)
    t_values = params / std_error
    p_values = 2 * (1 - t.cdf(np.abs(t_values), n_degree))

    return pd.DataFrame({
        'params': params,
        'std_error': std_error,
        't_values': t_values,
        'p_values': p_values,
    })



try:
    from pyspark.sql.window import Window
    from pyspark.sql import functions as F
    from pyspark.sql.types import StringType, DoubleType
    from pyspark.sql import DataFrame as TypeSparkDataFrame
except ImportError:
    print("pyspark not installed, please install it first. < pip install pyspark >")
from typing import List, Optional
from pandas import DataFrame as TypePandasDataFrame


@F.udf(returnType=DoubleType())
def udf_operator_div(num, den):
    """
    描述：用于 spark 中做除法，并区分分子分母状态，计算逻辑如下：

    分母\分子	   null	    0	   正值或负值
    null	  -4404	   -4404	-9999
    0	      -4404	   -4404	-9999
    正值或负值  -1111	    0	    正常计算

    参数：
    :param num: 分子名
    :param den: 分母名称

    Example:
    >>> data = [(None, 1), (1, 1), (None, None), (0, 0), (0, 1), (0, None), (1, None), (1, 0)]
    >>> df = spark.createDataFrame(data, schema=['num', 'den'])
    >>> df = df.withColumn('div', udf_operator_div(F.col('num'), F.col('den')))
    >>> df.show()
    +----+----+-------+
    | num| den|    div|
    +----+----+-------+
    |null|   1|-1111.0|
    |   1|   1|    1.0|
    |null|null|-4404.0|
    |   0|   0|-4404.0|
    |   0|   1|    0.0|
    |   0|null|-4404.0|
    |   1|null|-9999.0|
    |   1|   0|-9999.0|
    +----+----+-------+
    """
    if ((num is None) or (num == 0)) and ((den is None) or (den == 0)):
        return -4404.
    elif ((den is None) or (den == 0)):
        return -9999.
    elif num is None:
        return -1111.
    else:
        return num / den


def operator_div(num, den):
    """
    描述：用于 spark 中做除法，并区分分子分母状态，计算逻辑如下：

    分母\分子	   null	    0	   正值或负值
    null	  -4404	   -4404	-9999
    0	      -4404	   -4404	-9999
    正值或负值  -1111	    0	    正常计算

    参数：
    :param num: 分子名
    :param den: 分母名称

    示例：
    >>> data = [(None, 1), (1, 1), (None, None), (0, 0), (0, 1), (0, None), (1, None), (1, 0)]
    >>> df = spark.createDataFrame(data, schema=['num', 'den'])
    >>> df = df.withColumn('div', operator_div('num', 'den'))
    >>> df.show()
    +----+----+-------+
    | num| den|    div|
    +----+----+-------+
    |null|   1|-1111.0|
    |   1|   1|    1.0|
    |null|null|-4404.0|
    |   0|   0|-4404.0|
    |   0|   1|    0.0|
    |   0|null|-4404.0|
    |   1|null|-9999.0|
    |   1|   0|-9999.0|
    +----+----+-------+
    """
    cond_a = (((F.col(num).isNull()) | (F.col(num) == 0)) & ((F.col(den).isNull()) | (F.col(den) == 0)))
    cond_b = ((F.col(den).isNull()) | (F.col(den) == 0))
    cond_c = F.col(num).isNull()
    return F.when(cond_a, -4404).when(cond_b, -9999).when(cond_c, -1111).otherwise(F.col(num) / F.col(den))


# @F.udf(returnType=StringType())
def get_boundary(base_value, e=1):
    """
    描述：数值的区间脱敏，将数据转化为区间。

    参数：
    :param base_value: 区间脱敏
    :param e: 脱敏幂次

    示例：
    >>> df_sample = df_sample.withColumn('income', get_boundary(1000)(df_sample.income))
    1000为最低区分区间。
    """

    def f(value):
        try:
            value = float(value)
        except:
            return None

        if 'inf' in str(value):
            return str(value)

        # 正负值掩码
        mask = 1 if value >= 0 else -1

        # base_value 以内不进行分段区分
        value = value * mask
        if value < base_value and mask > 0:
            return f'[0, {base_value})'
        if value < base_value and mask < 0:
            return f'[-{base_value}, 0)'

        p = len(str(int(value))) - e

        low = int(value / float(f'1e{p}'))
        if mask >= 0:
            res = '[{}, {})'.format(str(int(mask * float(f'{low}e{p}'))), str(int(mask * float(f'{low + 1}e{p}'))))
        else:
            if value == float(f'{low}e{p}'):
                res = '[{1}, {0})'.format(str(int(mask * float(f'{low - 1}e{p}'))),
                                          str(int(mask * float(f'{low}e{p}'))))
            else:
                res = '[{1}, {0})'.format(str(int(mask * float(f'{low}e{p}'))),
                                          str(int(mask * float(f'{low + 1}e{p}'))))
        return res

    udf_fun = F.udf(f, returnType=StringType())
    return udf_fun



from pydantic import BaseModel, Field
from typing import List, Dict, Optional


__all__ = [
    "ROCCurveOutput",
    "PrecisionRecallCurveOutput",
    "ConfusionMatrixOutput",
    "BinaryMetricsOutput",
]


class ConfusionMatrixOutput(BaseModel):
    """"""
    tp: Optional[int] = Field(default=None, description="")
    fp: Optional[int] = Field(default=None, description="")
    tn: Optional[int] = Field(default=None, description="")
    fn: Optional[int] = Field(default=None, description="")


class ROCCurveOutput(BaseModel):
    """"""
    fpr: Optional[List[float]] = Field(default=None, description="")
    tpr: Optional[List[float]] = Field(default=None, description="")
    thresholds: Optional[List[float]] = Field(default=None, description="")


class PrecisionRecallCurveOutput(BaseModel):
    """"""
    recall: Optional[List[float]] = Field(default=None, description="")
    precision: Optional[List[float]] = Field(default=None, description="")
    thresholds: Optional[List[float]] = Field(default=None, description="")


class BinaryMetricsOutput(BaseModel):
    """"""
    ks: Optional[float] = Field(default=None, description="")
    auc: Optional[float] = Field(default=None, description="")
    recall: Optional[float] = Field(default=None, description="")
    lift: Optional[float] = Field(default=None, description="")
    precision: Optional[float] = Field(default=None, description="")
    f1_score: Optional[float] = Field(default=None, description="")
    confusion_matrix: Optional[ConfusionMatrixOutput] = Field(default=None, description="")
    roc_curve: Optional[ROCCurveOutput] = Field(default=None, description="")
    pr_curve: Optional[PrecisionRecallCurveOutput] = Field(default=None, description="")

    def show(self):
        """"""
        for key, value in self.model_dump().items():
            if key not in ["fpr", "tpr"]:
                if isinstance(value, float):
                    print(f"{key}: {value:.4f}")

    def to_univariate(self) -> Dict:
        """"""
        _eval = {
            "auc": self.auc,
            "ks": self.ks,
            "tpr": self.roc_curve.tpr,
            "fpr": self.roc_curve.fpr,
        }
        return _eval



try:
    from pyspark.sql import functions as F
    from pyspark.sql.types import StringType
    from pyspark.sql import DataFrame as TypeSparkDataFrame
except ImportError:
    print("pyspark not installed, please install it first. < pip install pyspark >")
from typing import Any, List, Dict


__all__ = [
    "convert_col_to_list",
    "map_label",
    "get_table_columns"
]


def convert_col_to_list(df: TypeSparkDataFrame, column: str) -> List[Any]:
    """
    desc: convert pyspark DataFrame to list.

    :param df: spark dataframe.
    :param col: column name.
    :return: column value list.
    """
    return df.select(column).rdd.map(lambda x: x[0]).collect()


def map_label(mapping: Dict):
    """
    描述：[UDF] 用来匹配 pyspark DataFrame 中的键值对

    示例：使用行业分类码值表，设置一列中文行业名称。
    >>> df.withColumn('industryCnName', mapLabel(industryDict)(df.industryCode))
    """
    def f(value):
        return mapping.get(value)
    return F.udf(f, returnType=StringType())


def get_table_columns(db_name: str, table_name: str, spark: TypeSparkDataFrame) -> List[str]:
    """
    描述：获取表的列名称
    示例：
    >>> getTableColumns('temp_work', 'test', spark)
    """
    columns_info = spark.catalog.listColumns(table_name, db_name)
    columns_name = [col.name for col in columns_info]
    return columns_name



import pandas as pd
from typing import List, Tuple, Optional
from pyspark.sql import SparkSession
from pandas import DataFrame as TypePandasDataFrame

__all__ = [
    "show_table_partition",
    "get_index_partition",
    "drop_partition_data",
]


def show_table_partition(
        db_name: Optional[str] = None,
        table_name: Optional[str] = None,
        spark: Optional[SparkSession] = None,
) -> TypePandasDataFrame:
    """
    描述：展示表的分区信息
    Example:
        df_part = ShowTablePartitions(table_name='tmp_work.t_cch_sample_long', spark=spark)
    """
    if db_name and table_name:
        table_name = f'{db_name}.{table_name}'

    def parse_partition_value(row_value):
        row = dict()
        for item in row_value:
            key, value = item.split('=')
            row.update({key: value})
        return row
    df_part = spark.sql(f"SHOW PARTITIONS {table_name}").toPandas()
    part = df_part.partition.str.split('/').tolist()
    return pd.DataFrame(list(map(parse_partition_value, part)))


def get_index_partition(
        db_name: str,
        table_name: str,
        spark: SparkSession
) -> List[Tuple[str, str]]:
    """

    :param db_name:
    :param table_name:
    :param spark:
    :return:
    """
    df_part = show_table_partition(db_name, table_name, spark=spark)
    if len(df_part) == 0:
        return []
    else:
        return list(zip(df_part.load_date, df_part.index_name))


def drop_partition_data(db_name: str, table_name: str, partition: str, spark: SparkSession) -> None:
    """
    描述：删除分区 示例：
    """
    query = f"""
        alter table {db_name}.{table_name} drop if exists partition (partition)
        """
    spark.sql(query)
    print(f'Table {db_name}.{table_name} partition: {partition} has been dropped.')



from numpy.typing import NDArray
from typing import Optional
from pydantic import BaseModel, Field, ConfigDict


__all__ = ["MakeDataOutput"]


class MakeDataOutput(BaseModel):
    """"""
    model_config = ConfigDict(arbitrary_types_allowed=True)
    x_train: Optional[NDArray] = Field(default=None, description="")
    y_train: Optional[NDArray] = Field(default=None, description="")
    x_test: Optional[NDArray] = Field(default=None, description="")
    y_test: Optional[NDArray] = Field(default=None, description="")



from tqdm import tqdm
from ..spark import unpivot, insert_into_hive, get_index_partition
from ..utils.date import date_range
from ..spark.operator import operator_mapping
from pyspark.sql import functions as F
from pyspark.sql.types import *
from ..spark import pct_rank


__all__ = [
    'one_batch_data',
    'features_agg',
    'batch_unpivot',
    'gen_batch_features',
]


def one_batch_data(df, now_date, month_range=[3, 6, 9, 12], fmt: str = 'yyyyMMdd'):
    """"""
    df = df.withColumn('now_date', F.lit(now_date))
    df = df.withColumn('month_diff', F.months_between(F.to_date('now_date', fmt), F.to_date('load_date', fmt)))

    batch_datas = [df.where(F.col('month_diff').between(0, month - 1)) for month in month_range]
    return batch_datas


def features_agg(batch_datas, column='info', month_range=[3, 6, 9, 12], operator=['cnt', 'sum', 'avg']):
    """

    :param batch_datas:
    :param column:
    :param month_range:
    :param operator: 计算算子，默认为['cnt', 'sum', 'avg']。支持以下算子：
        - 'cnt'
        - 'avg'
        - 'sum'
        - 'cnt_dist'
        - 'sum_dist'
        - "min"
        - "max"
        - "mean"
        - "first"
        - "last"
        - "skewness"
        - "kurtosis"
        - "collect_set"
        - "collect_list"
        - "approx_count_distinct"
        - 'stddev'
        - 'stddev_samp'
        - 'stddev_pop'
        - 'variance'
        - 'var_pop'
        - 'var_samp'
    :return:

    示例：
    >>> load_date = '20221130'
    >>> df_features_zzc = gen_batch_features(
    >>>     df_origin, column='zzc', now_date=load_date, month_range=[3, 6, 9, 12],
    >>>     operator=['cnt', 'sum'], spark=spark)
    >>> df_features_zzc.show()
    >>> df_origin
    +----+---+---+---------+
    |uuid|zzc|nsr|load_date|
    +----+---+---+---------+
    |   3| 81| 10| 20191231|
    |   3| 40| 73| 20200131|
    |   0| 24| 44| 20200229|
    |   2| 79| 75| 20200331|
    |   1| 22| 37| 20200430|
    |   3|  0| 10| 20200531|
    """
    agg_datas = []
    for batch_data, month in zip(batch_datas, month_range):

        op = [operator_mapping.get(item)(column).alias(f'ent_{column}_last_{month}_{item}') for item in operator]

        agg_data = batch_data.groupBy('uuid').agg(*op)
        agg_datas.append(agg_data)
    return agg_datas


def batch_unpivot(agg_features, now_date='20221130', spark=None):
    """"""
    schema = StructType([
        StructField("uuid", StringType(), True),
        StructField("index_name", StringType(), True),
        StructField("index_val", FloatType(), True),
    ])

    df = spark.createDataFrame([], schema=schema)

    for batch_feature in agg_features:
        columns = batch_feature.drop('uuid').columns
        df_batch = unpivot(
            batch_feature, columns, val_type=FloatType, index_name='uuid',
            feature_name='index_name', feature_value='index_val')
        df = df.unionAll(df_batch)

    df = df.withColumn('index_id', F.lit(''))
    df = df.withColumn('load_date', F.lit(now_date))
    return df


def gen_batch_features(
        df_origin, column='info', now_date='20221130',
        month_range=[3, 6, 9, 12], operator=['cnt', 'sum', 'avg'],
        spark=None):
    """
    描述：生成一个时间点的回溯跑批
    :param df_origin:
    :param column:
    :param now_date:
    :param month_range:
    :param operator: 计算算子，默认为['cnt', 'sum', 'avg']。
    :param spark:
    :return:
    """
    feature_datas = one_batch_data(df_origin, now_date=now_date, month_range=month_range)
    agg_features = features_agg(feature_datas, column=column, month_range=month_range, operator=operator)
    df_feature = batch_unpivot(agg_features, now_date=now_date, spark=spark)
    return df_feature


def batch_derivative_index(
        df_origin, column=None, start_date=None, end_date=None,
        db_name=None, table_name=None, insert=True, spark=None):
    """

    :param df_origin:
    :param column:
    :param start_date:
    :param end_date:
    :param insert:
    :param spark:
    :return:
    """
    dates = date_range(start_date, end_date)

    for load_date in tqdm(dates, desc=column):
        df_features = gen_batch_features(
            df_origin, column=column, now_date=load_date, month_range=[3, 6, 9, 12], spark=spark)

        index_names = df_features.select('index_name').distinct().collect()
        index_names = [item['index_name'] for item in index_names]

        df_features = df_features.select('uuid', 'index_val', 'index_id').persist()

        for index_name in index_names:

            df_features = df_features.where(F.col('index_name') == index_name)
            if df_features.count() > 0:
                if insert:
                    insert_into_hive(
                        df_features, db_name=db_name, table_name=table_name, spark=spark,
                        partition=['index_name', 'load_date'],
                        partition_val=[index_name, load_date], )
                else:
                    df_features.show()
            else:
                pass
        df_features.unpersist()

    return None


def index_indu_pct(db_name, table_name, spark, sqlContext, part_info=None, industry_level='l1_code', tmp_table_name='temp_table'):
    """
    描述：行业排名衍生
    :param db_name: 数据库名
    :param table_name: 数据表名，原始数据和新生成的行业排名指标将储存在同一库中。
    :param spark: spark session
    :param part_info: 跑批信息，[(load_date, column_name), (load_date, column_name)]
    :param industry_level: 行业级别，默认为`l1_code`。
    :param tmp_table_name: 临时表表名称，默认为`temp_table`。
    :return: None
    """

    industry_col = [
        'industry_l1_code','industry_l2_code','industry_l3_code','industry_l4_code',]

    if part_info is None:
        part_info = get_index_partition(db_name, table_name, spark)
    else:
        pass

    for load_date, col in part_info:
        pct_col = f'rank_indu_{col}'
        if ((load_date, pct_col) not in part_info) and (col not in industry_col) and ('rank_indu'not in col):
            df_index = spark.sql(f"""
                SELECT uuid, index_val FROM {db_name}.{table_name}
                WHERE load_date ='{load_date}'
                    AND index_val IS NOT NULL
                    AND index_name '{col}'
                """)
            df_index = df_index.where(F.col('index_val').isNotNull() & ~F.isnan('index_val'))
            df_index.write.mode("overwrite").parquet(tmp_table_name)
            df_index = sqlContext.read.parquet(tmp_table_name)
            df_index = df_index.persist()

            df_indu = spark.sql(f"""
                SELECT uuid, index_val AS indu
                FROM {db_name}.t_ent_index_basic
                WHERE load_date = '{load_date}'
                    AND index_name = 'industry_{industry_level}'
                """)
            if table_name == 't_ent_index_basic':
                df_indu.write.mode("overwrite").parquet('tmp_indu')
                df_indu = sqlContext.read.parquet('tmp_indu')

            df_indu = df_indu.persist()

            if df_index.count() > 0:
                df_index = df_index.join(df_indu, on='uuid', how='left')
                df_index = pct_rank(df_index, 'index_val', index_name=col, partitionby=['indu'])

                # return df_index,Load_date,pct_col,db_name,table_name
                df_index = df_index.select('uuid', F.col(pct_col).alias('index_val'))
                insert_into_hive(
                    df_index, db_name, table_name, spark=spark,
                    partition=['load_date', 'index_name'],
                    partition_val=[load_date, pct_col])
                df_index.unpersist()
                df_indu.unpersist()
            else:
                pass
        else:
            pass



import numpy as np
import pandas as pd
try:
    from sklearn.cluster import KMeans
except ImportError:
    raise ImportError("Please install scikit-learn to use this module. pip install scikit-learn")
from pandas import DataFrame, Series
from typing import Any, List, Union, Optional
from kivi.woe.base import WOEMixin


__all__ = ['KmeansBins']


class KmeansBins(WOEMixin):
    """ 描述：Kmeans 分箱方法，使用 Kmeans 进行分箱分析。 """
    cutoff_points: List[float]

    def __init__(
            self,
            variables: Series,
            target: Series,
            bins: Optional[int] = 5,
            n_init: Optional[int] = 10,
            random_state: Optional[int] = 0,
            abnormal_vals: Optional[List[Union[str, int, float]]] = None,
            fill_bin: Optional[bool] = True,
            decimal: Optional[int] = 6,
            weight: Optional[Any] = None,
            *args: Any,
            **kwargs: Any,
    ):
        """
        描述：[无监督分箱] Kmeans 分箱方法，使用 Kmeans 进行分箱分析。

        :param variables: 待分箱变量
        :param target: 目标标签变量
        :param bins: 决策树分箱中最大的叶子结点数量，一般对应的是最终分箱数量，默认为 5 。
        :param abnormal_vals: 特殊值分箱，在变量存在特殊值时单独分一箱，如 -1111, -9999。
        :param fill_bin: 在各分箱中偶发性会出现 good 或 bad 为 0 的情况，默认 fill_pos 为 True ，为该分箱填充 0.5。

        Example:
            woe = KmeansBins(variables, target, bins=5, fill_bin=True)
            woe.fit()
        """
        self.variables = variables
        self.target = target
        self.bins = bins
        self.n_init = n_init
        self.random_state = random_state
        self.fill_bin = fill_bin
        self.abnormal_vals = abnormal_vals
        self.decimal = decimal
        self.args = args
        self.kwargs = kwargs
        self.data_prepare(
            variables=variables, target=target, weight=weight)

    def kmeans_cutoff_point(self,):
        """"""
        vars = self.variables.to_numpy().reshape(-1, 1)
        kmeans = KMeans(n_init=self.n_init, n_clusters=self.bins, random_state=self.random_state).fit(vars)
        df_var_label = pd.DataFrame({
            'label': kmeans.fit_predict(vars).flatten(),
            'vars': self.variables
        })
        bins = sorted(df_var_label.groupby('label').max().vars.tolist())[:-1]
        bins += [np.inf, -np.inf]
        bins.sort()
        self.cutoff_point = bins

    def fit(
            self,
            score: Optional[bool] = True,
            origin_border: Optional[bool] = False,
            order: Optional[bool] = True,
            **kwargs: Any,
    ) -> DataFrame:
        """

        :param score: 是否增加 WOEMixin score。
        :param origin_border: 是否增加 分箱中的最大值与最小值。
        :param order: 是否增加单调性判断。
        :return: DataFrame WOEMixin result.
        """
        self.kmeans_cutoff_point()
        _bucket, _bins = pd.cut(
            self.df_data.variables, self.cutoff_point, include_lowest=True, retbins=True)
        bucket = pd.DataFrame({
            'variables': self.df_data.variables,
            'target': self.df_data.target,
            'bucket': _bucket,
        }).groupby('bucket', as_index=True, observed=False)
        self.cal_woe_iv(bucket, score=score, origin_border=origin_border, order=order)
        return self.woe







from itertools import product, combinations
from collections import defaultdict
import statsmodels.api as sm
from tqdm import tqdm
import pandas as pd
from ..ModelTools import get_ks

from sklearn.linear_model import LogisticRegressionCV
from ..ModelTools import statsmodel_metrics, sklearn_metrics

class Logistic:

    def __init__(self, data, y_name, alphas=None, framework='sklearn', threshold=0.5):

        y = data.pop(y_name)
        self.exog = data.copy()
        self.endog = y.copy()
        self.framework = framework
        self.threshold=threshold


    def logistic(self, exog, endog):

        if self.framework == 'sklearn':
            model = LogisticRegressionCV(cv=5, random_state=0)
            self.model = model.fit(exog, endog)

        elif self.framework == 'statsmodel':
            self.model = sm.Logit(endog, exog).fit(disp=0)

        else:
            Exception(ValueError, 'Framework Error!')


    def metrics(self, model, col, res, prefix):
        if self.framework=='statsmodel':
            statsmodel_metrics(
                res=res,
                prefix=prefix,
                model=model,
                col=col,
                threshold=self.threshold,
            )
        elif self.framework=='sklearn':
            sklearn_metrics(
                res=res,
                model=model,
                threshold=self.threshold,
                prefix=prefix,
            )


    def plot_roc(self):
        pass


    def vif(self, exog, exog_idx=None, ls=False):
        """
        Des: 方差膨胀因子 - 多重共线性参考指标
        :param exog: 变量
        :param exog_idx: 变量索引，在不指定变量索引时计算全部变量的vif，并以字典形式返回计算结果
        :param ls: 以list的形式返回数据
        :return: float or dict vif
        """
        from statsmodels.stats.outliers_influence import variance_inflation_factor
        if exog_idx:
            return variance_inflation_factor(exog, exog_idx)
        elif ls:
            return [variance_inflation_factor(exog.values, i) for i in range(exog.shape[-1])]
        else:
            vif_dict = {}
            for i, name in enumerate(exog):
                if name == "const":
                    continue
                vif_dict[name] = variance_inflation_factor(exog.values, i)
            return vif_dict


    def different_variable_combinations(self, col_description=None, normalized=True, *args, **kwargs):
        """
        Des: OLS Different variable combinations
        :param features_num: 变量数
        :return: model_index, variables_index
        """
        res_dict = defaultdict(list)
        cols = self.exog.columns.to_list()

        variables_index = pd.DataFrame()

        if "const" in cols:
            cols.remove("const")

        if isinstance(args[0], int) and len(args) == 1:
            if len(cols) < args[0]:
                raise Exception(ValueError, 'features_num must more than the columns!')
            combins = list(enumerate(combinations(cols, args[0])))

        else:
            combins = list(product(*args))

        if combins:
            for temp_cols in tqdm(combins, desc=f'Total regressions {len(combins)}'):
                temp_cols = list(temp_cols)
                temp_data = self.exog[temp_cols]

                # data clean
                temp_data['target'] = self.endog
                temp_data.dropna(axis=0, inplace=True)

                if temp_data.target.sum() != 0:
                    res_dict['coll_name'].append(temp_cols)

                    # chinese description for columns
                    if isinstance(col_description, dict):
                        if len(temp_cols) > 1:
                            res_dict['des'].append([col_description.get(col) for col in temp_cols])
                        else:
                            res_dict['des'].append(col_description.get(temp_cols))
                            try:
                                res_dict['ks'].append(get_ks(temp_data[temp_cols], temp_data.target)[0])
                            except:
                                res_dict['ks'].append(0)

                    res_dict['missing_rate'].append(float(len(temp_data) / len(self.endog)))
                    res_dict['bad'].append(temp_data.target.sum())
                    res_dict['count'].append(temp_data.target.count())
                    res_dict['good'].append(temp_data.target.count() - temp_data.target.sum())
                    res_dict['bad_rate'].append(temp_data.target.sum() / temp_data.target.count())

                    X = sm.add_constant(temp_data[temp_cols])
                    try:
                        self.logistic(X, temp_data.target)
                    except Exception as e:
                        print(e, temp_cols)

                    statsmodel_metrics(res=res_dict, prefix='', model=self.model, col=temp_cols)

                    # if kwargs.get('normalized'):
                    if normalized:
                        X = sm.add_constant(
                            (temp_data[temp_cols] - temp_data[temp_cols].min()) / (temp_data[temp_cols].max() - temp_data[temp_cols].min()),
                        )

                        try:
                            self.logistic(X, temp_data.target)
                        except Exception as e:
                            print(e, temp_cols)

                        statsmodel_metrics(res=res_dict, prefix='normal_', model=self.model, col=temp_cols)

                    if len(temp_cols) != 1:
                        vifLs = self.vif(temp_data[temp_cols], ls=True)
                    res_dict['vif'].append(vifLs)
                else:
                    continue

            return pd.DataFrame(res_dict)
        else:
            raise Exception(ValueError, 'combins are None!')




from pyspark.ml.feature import VectorAssembler
from pyspark.ml.classification import LogisticRegression
from pyspark.ml.regression import LinearRegression
from pyspark.ml.stat import ChiSquareTest


__all__ = ['SparkUnivariate']


class SparkUnivariate(object):
    """

    """
    def __init__(self, df, target_name='target', na='drop'):
        self.df = df

    def describe(self, ):
        df_res = self.df.describe()

    def modelRes(self, col):
        return None



from typing import Union, Optional
import pandas as pd
from pyspark.sql import functions as F
from pyspark.sql.types import DoubleType


__all__ = [
    "merge_index_name",
    "operator_mapping",
    "operator_growth_months",
    "operator_consecutive_growth_months",
    "operator_pandas_udf_group_sum_ratio",
]


def merge_index_name(
        id_level: str,
        index_name: str,
        operate_name: str,
        month: Union[str, int],
        alias_operator: Optional[str] = None,
) -> str:
    """
    desc: 合并指标名称

    :param id_level: 聚合主键
    :param index_name: 指标名称
    :param operate_name: 聚合算子
    :param month: 月份
    :param alias_operator: 别名算子
    :return: 指标名称
    """
    if alias_operator:
        return '_'.join([id_level, index_name, alias_operator, f'{month}m'])
    else:
        return '_'.join([id_level, index_name, operate_name, f'{month}m'])


def operator_ratio(num, den):
    """"""
    if num is None and den is None:
        return -4404.
    elif num is None:
        return -1111.
    elif den is None or den == 0.:
        return -9999.
    else:
        return num / den


def operator_growth_months(data):
    """"""
    consecutive_months = 0
    for i in range(1, len(data)):
        if data[i] > data[i - 1]:
            consecutive_months += 1
    return consecutive_months


def operator_consecutive_growth_months(data):
    """"""
    max_growth = 0
    current_growth = 0
    for pre, aft, in zip(data[:-1], data[1:]):
        if aft > pre:
            current_growth += 1
        else:
            current_growth = 0
        max_growth = max(max_growth, current_growth)
    return max_growth


def operator_pandas_udf_group_sum_ratio(num: pd.Series, den: pd.Series) -> float:
    """"""
    _num = num.sum()
    _den = den.sum()
    if _den == 0.:
        return -9999.
    else:
        return _num / _den


operator_mapping = {
    "sum": F.sum,
    "avg": F.avg,
    "mean": F.mean,
    "ratio": F.udf(f=operator_ratio, returnType=DoubleType()),
    "group_sum_ratio": F.pandas_udf(f=operator_pandas_udf_group_sum_ratio, returnType=DoubleType()),
}
index_schema_list = ['uuid', 'index_name', 'index_val', 'load_date']



import os
import numpy as np
import pandas as pd
from pandas import DataFrame
from typing import Tuple, Optional
from sklearn import datasets
from sklearn.model_selection import train_test_split
from .schema import *


__all__ = [
    "Dataset",
    "MakeData",
]
path = os.path.dirname(__file__)


class Dataset:
    """
    描述：返回用于示例文档的数据集。

    Example:
        dataset = Dataset()
        df_bank = dataset.bank_data()
        df_crime = dataset.crime_data()
    """
    @staticmethod
    def bank_data():
        """
        描述：获取一份信贷数据，包括个人主体信息以及违约标签。
        :return: DataFrame
        """
        filename = os.path.join(path, 'bank.xlsx')
        df = pd.read_excel(filename)
        df['target'] = df['y'].apply(lambda x: 1 if x == 'yes' else 0)
        return df.drop('y', axis=1)

    @staticmethod
    def crime_data():
        """
        描述：获取一份欺诈犯罪数据，包括个人主体信息以及欺诈标签。
        :return: DataFrame
        """
        filename = os.path.join(path, 'crime_data.csv')
        return pd.read_csv(filename)


class MakeData:

    def __init__(
            self,
            n_samples: Optional[int] = 10000,
            n_features: Optional[int] = 5,
            n_informative: Optional[int] = 5,
            n_redundant: Optional[int] = 0,
            test_size: Optional[float] = 0.3,
    ):
        self.n_samples = n_samples
        self.n_features = n_features
        self.n_informative = n_informative
        self.n_redundant = n_redundant
        self.test_size = test_size

    def dataset(self, ) -> MakeDataOutput:
        np.random.seed(0)
        x, y = datasets.make_classification(
            n_samples=self.n_samples, n_features=self.n_features,
            n_informative=self.n_informative, n_redundant=self.n_redundant)
        x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=self.test_size, random_state=42)
        data = MakeDataOutput(x_train=x_train, x_test=x_test, y_train=y_train, y_test=y_test)
        return data

    def sample(self) -> Tuple[DataFrame, DataFrame]:
        """"""
        columns_name = [f"col_{i}" for i in range(self.n_features)]
        data = self.dataset()
        train = pd.concat([pd.DataFrame(data.x_train, columns=columns_name), pd.DataFrame(data.y_train, columns=['target'])], axis=1)
        test = pd.concat([pd.DataFrame(data.x_test, columns=columns_name), pd.DataFrame(data.y_test, columns=['target'])], axis=1)
        train["uuid"] = np.arange(len(train))
        test["uuid"] = np.arange(len(test))
        return train, test



import pandas as pd
import statsmodels.api as sm
from pandas import DataFrame
from typing import Any, List, Callable, Literal, Optional

from ..utils.logger import LoggerMixin
from ..evaluate import BinaryMetrics
from .schema import StepwiseReport


__all__ = [
    "StepWise",
    "StepwiseMixin",
]


class StepwiseMixin(LoggerMixin):
    """逐步回归法"""
    df: DataFrame

    def implication(self, model, columns: List[str], report: StepwiseReport) -> StepwiseReport:
        """"""
        params = model.params[columns]
        report.params = params.round(6).to_dict()
        report.woe_impact = params[params < 0].index.tolist()
        report.positive_impact = params[params > 0].index.tolist()
        report.negative_impact = params[params < 0].index.tolist()
        return report

    def step_fit(self, columns: List[str], target, report: StepwiseReport) -> StepwiseReport:
        """"""
        x_data = sm.add_constant(self.df[columns], has_constant='add')
        y_data = target
        report.num_features = len(columns)
        try:
            model = sm.Logit(y_data, x_data).fit(disp=0)
            report = self.implication(model, columns, report)
            binary_metrics = BinaryMetrics(target=model.model.endog, proba=model.predict())
            metrics = binary_metrics.evaluate()
            report = report.model_copy(update=metrics.model_dump())
        except Exception as e:
            self._logger(msg=f'Model Error! Feature Name: {columns}, Error Msg: {e}', color="red")
        return report


class StepWise(StepwiseMixin):
    """向前逐步回归"""
    def __init__(
            self,
            df: pd.DataFrame,
            columns: Optional[List[str]] = None,
            target_name: Optional[str] = "target",
            var_field: Optional[str] = "var_name",
            dependent: Optional[DataFrame] = None,
            indicator: Optional[Literal["auc", "ks", "lift"]] = 'auc',
            early_stop: Optional[float] = None,
            logger: Optional[Callable] = None,
            verbose: Optional[bool] = True,
            *args: Any,
            **kwargs: Any,
    ):
        """"""
        self.df = df
        self.target_name = target_name
        self.var_field = var_field
        self.columns = columns
        self.dependent = dependent
        self.indicator = indicator
        self.early_stop = early_stop
        self.logger = logger
        self.verbose = verbose
        self.args = args
        self.kwargs = kwargs
        self._validate_data()
        self._prepare_data()

    def __call__(self, step_type: Literal["forward", "backward"], *args, **kwargs) -> DataFrame:
        """"""
        if step_type == "forward":
            return self.forward()
        elif step_type == "backward":
            return self.backward()
        else:
            raise ValueError(f"step_type {step_type} not supported")

    def _validate_data(self):
        """"""
        if self.target_name not in self.df.columns:
            raise ValueError(f"target_name {self.target_name} not in df.columns: {self.df.columns.tolist()}")
        if self.var_field not in self.dependent.columns:
            raise ValueError(f"var_field {self.var_field} not in dependent.columns: {self.dependent.columns.tolist()}")

    def _validate_early_stop(
            self,
            steps_type: Literal["forward", "backward"] = "forward"
    ):
        """"""
        if steps_type == "forward" and self.early_stop < 0:
            raise ValueError(f"early_stop increase {self.early_stop} must be greater than or equal to 0")

        if steps_type == "backward" and self.early_stop > 0:
            raise ValueError(f"early_stop decrease {self.early_stop} must be less than or equal to 0")

    def _prepare_data(self):
        if self.columns is None:
            self.columns = self.df.columns.tolist()

    def _get_candidate_columns(self):
        """"""
        self.candidate_columns = self.dependent.sort_values(by=str(self.indicator), ascending=False)[self.var_field].tolist()
        self.candidate_columns = [column for column in self.candidate_columns if column in self.columns]

    def _get_step_columns(
            self,
            i_step: int,
            steps_type: Literal["forward", "backward"] = "forward"
    ) -> List[str]:
        """"""
        if steps_type == "forward":
            columns = self.candidate_columns[: i_step + 1]
        else:
            if i_step == 0:
                columns = self.candidate_columns
            else:
                columns = self.candidate_columns[:-i_step]
        return columns

    def early_stop_monitor(
            self,
            steps: List[StepwiseReport],
            i_step: int,
            steps_type: Literal["forward", "backward"] = "forward",
    ) -> bool:
        """"""
        if isinstance(self.early_stop, float) and i_step > 0:
            pre_step = steps[i_step - 1]
            cur_step = steps[i_step]
            pre_indicator = pre_step.model_dump().get(str(self.indicator))
            cur_indicator = cur_step.model_dump().get(str(self.indicator))
            indicator_change = (cur_indicator - pre_indicator) / pre_indicator
            cur_step.indicator_change = indicator_change
            if indicator_change < self.early_stop:
                if steps_type == "forward":
                    self._logger(
                        msg=f"dependent {self.indicator} increase is {indicator_change:.4f} < {self.early_stop}, stop.",
                        color="blue")
                elif steps_type == "backward":
                    self._logger(
                        msg=f"dependent {self.indicator} decrease is {indicator_change:.4f} < {self.early_stop}, stop.",
                        color="blue")
                return True
            else:
                return False
        else:
            return False

    def steps(
            self,
            steps_type: Literal["forward", "backward"] = "forward"
    ) -> DataFrame:
        """"""
        self._get_candidate_columns()
        steps: List[StepwiseReport] = []

        if isinstance(self.early_stop, float):
            self._validate_early_stop(steps_type=steps_type)

        for i in range(len(self.candidate_columns)):
            columns = self._get_step_columns(i_step=i, steps_type=steps_type)
            step_report = StepwiseReport(feature=columns)
            step_report = self.step_fit(columns=columns, target=self.df[self.target_name], report=step_report)
            steps.append(step_report)
            if self.early_stop_monitor(steps=steps, i_step=i, steps_type=steps_type):
                break
        return pd.DataFrame([step.model_dump() for step in steps])

    def forward(self) -> DataFrame:
        """"""
        return self.steps("forward")

    def backward(self) -> DataFrame:
        """"""
        return self.steps("backward")

    @staticmethod
    def top_n_features(df_woe, var_name='var_name', sort_by='ks', top_n=3, dim='dim'):
        """
        描述：依据指标的效果排序筛选指标。
        :param df_woe:
        :param var_name:
        :param sort_by:
        :param top_n:
        :param dim:
        :return:
        """
        columns = [var_name, 'des', 'ks', 'auc', 'dim']
        df_features = pd.DataFrame(columns=columns)
        df_woe = df_woe[columns].drop_duplicates()
        for group_name, group in df_woe.groupby(dim):
            df_group = group.copy()
            df_group.sort_values(sort_by, ascending=False, inplace=True)
            df_features = df_features.append(df_group.head(top_n))
        return df_features





from sklearn.feature_extraction.text import CountVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.metrics import jaccard_score
from difflib import Differ, SequenceMatcher
import jieba
import logging
import warnings

warnings.filterwarnings('ignore')
jieba.setLogLevel(logging.INFO)

def check_contain_chinese(check_str):
    for ch in check_str:
        if u'\u4e00' <= ch <= u'\u9fff':
            return True
    return False


class similarity(object):
    def __init__(self):
        pass

    def transform(self, text1, text2):
        """
        把文本text1，text2转化为英文样式的text1，text2和向量vec1，vec2
        :param text1:
        :param text2:
        :return:
        """

        if check_contain_chinese(text1):
            text1 = ' '.join(jieba.lcut(text1))
            text2 = ' '.join(jieba.lcut(text2))
        else:
            pass

        corpus = [text1, text2]
        cv = CountVectorizer()
        cv.fit(corpus)
        vec1 = cv.transform([text1]).toarray()
        vec2 = cv.transform([text2]).toarray()
        return text1, text2, vec1, vec2

    def compute(self, text1, text2):
        """
        对输入的text1和text2进行相似性计算，返回相似性信息
        :param text1:  文本字符串
        :param text2: 文本字符串
        :return:  字典， 形如{
                'Sim_Cosine':0.8,
                'Sim_Jaccard': 0.3,
                'Sim_MinEdit': 0.5,
                'Sim_Simple': 0.8
                }
        """
        text1, text2, vec1, vec2 = self.transform(text1, text2)
        data = {
                'Sim_Cosine': self.cosine_similarity(vec1, vec2),
                'Sim_Jaccard': self.jaccard_similarity(vec1, vec2),
                'Sim_MinEdit': self.minedit_similarity(text1, text2),
                'Sim_Simple': self.simple_similarity(text1, text2)
                }
        return data

    def cosine_similarity(self, vec1, vec2):
        cos_sim = cosine_similarity(vec1, vec2)
        return cos_sim[0][0]


    def jaccard_similarity(self, vec1, vec2):
        """ returns the jaccard similarity between two lists """
        try:
            return jaccard_score(vec1[0], vec2[0])
        except:
            warnings.warn('scikit-learn版本问题，可以在命令行执行"pip install scikit-learn==0.22"')


    def minedit_similarity(self, text1, text2):
            words1 = jieba.lcut(text1.lower())
            words2 = jieba.lcut(text2.lower())
            leven_cost = 0
            s = SequenceMatcher(None, words1, words2)
            for tag, i1, i2, j1, j2 in s.get_opcodes():
                if tag == 'replace':
                    leven_cost += max(i2 - i1, j2 - j1)
                elif tag == 'insert':
                    leven_cost += (j2 - j1)
                elif tag == 'delete':
                    leven_cost += (i2 - i1)
            return leven_cost

    def simple_similarity(self, text1, text2):
        words1 = jieba.lcut(text1.lower())
        words2 = jieba.lcut(text2.lower())
        diff = Differ()
        diff_manipulate = list(diff.compare(words1, words2))
        c = len(diff_manipulate) / (len(words1) + len(words2))
        cmax = max([len(words1), len(words2)])
        return (cmax - c) / cmax




import numpy as np
from pyspark.sql import functions as F

__all__ = [
    "quantile",
    "getQuantiles",
    "psi",
    "csi",
]


def quantile(col: str, percent=0.5):
    """
    描述：分位数计算

    参数：
    :param col: 字段名称
    :param percent: 分位点
    :return : pyspark.sql.functions
    """
    return F.expr(f'percentile_approx({col}, {percent})')


def getQuantiles(df, column_name, ranges=np.linspace(0, 1, 11))->list:
    """
    描述：获取字段的分位数点。

    参数：
    :param df:
    :param column_name:
    :param ranges:
    :return : 分位数值 List

    Example：获取多个字段的多个分位点
        quant_val_list = [getQuantiles(df, col) for col in cols]
        pd.DataFrame(dict(zip(cols, quant_val_list)), index=np.linspace(0, 1, 11)).T
    """
    quantiles = [quantile(column_name, i).alias(f'i') for i in ranges]
    return df.agg(*quantiles).toPandas().values.flatten().tolist()


def psi(df_buckets, ex_name='Expected', ac_name='Actual', fill_bin=True):
    """

    :param df_buckets:
    :param ex_name:
    :param ac_name:
    :return:
    """
    if fill_bin:
        if 0 in df_buckets[ex_name].tolist():
            df_buckets[ex_name] += 0.5
        if 0 in df_buckets[ac_name].tolist():
            df_buckets[ac_name] += 0.5

    df_buckets['Expected_percentage'] = df_buckets[ex_name] / df_buckets[ex_name].sum()
    df_buckets['Actual_percentage'] = df_buckets[ac_name] / df_buckets[ac_name].sum()
    df_buckets['A-E'] = df_buckets.Actual_percentage - df_buckets.Expected_percentage
    df_buckets['A/E'] = df_buckets.Actual_percentage / df_buckets.Expected_percentage
    df_buckets['ln(A/E)'] = np.log(df_buckets['A/E'])
    df_buckets['psi_val'] = df_buckets['A-E'] * df_buckets['ln(A/E)']
    df_buckets['psi'] = df_buckets['psi_val'].sum()
    return df_buckets


def csi(df_buckets, ex_name='Expected', ac_name='Actual', fill_bin=True):
    """
    Des: 计算CSI的核心函数
    :param df_buckets: 明细表
    :return: 明细表以及CSI值 df_buckets, csi
    """

    df_buckets['Expected_percentage'] = df_buckets.Expected / df_buckets.Expected.sum()
    df_buckets['Actual_percentage'] = df_buckets.Actual / df_buckets.Actual.sum()
    df_buckets['A-E'] = df_buckets.Actual_percentage - df_buckets.Expected_percentage
    if len(df_buckets) != len(self.score):
        raise Exception(TypeError, 'Score len error!')
    df_buckets['score'] = self.score
    df_buckets['index'] = df_buckets['A-E'] * self.score
    df_buckets['csi'] = df_buckets['index'].sum()
    return df_buckets



import numpy as np
import pandas as pd
from pandas import DataFrame, Series
from typing import Any, List, Tuple, Union, Callable, Optional
import statsmodels.api as sm

from ..utils.operator import StatsLogit, NumRange
from ..utils.logger import LoggerMixin
from .psi import psi
from .utils import lift
from .binary_metrics import BinaryMetrics


__all__ = [
    "ScoreEvaluate",
    "ModelEvaluate",
]


class ScoreEvaluate(LoggerMixin):
    """"""
    df_lifts: DataFrame
    lifts: List[DataFrame]

    def __init__(
            self,
            scores: Union[DataFrame, List[DataFrame]],
            score_name: Optional[str] = 'score',
            target_name: Optional[str] = 'target',
            bins: Optional[Union[int, List[float]]] = 20,
            border: Optional[Tuple[Union[int, float], Union[int, float]]] = None,
            keys: Optional[List[str]] = None,
            logger: Optional[Callable] = None,
            verbose: Optional[bool] = False,
            **kwargs: Any
    ):
        """
        :param scores: DataFrame 评分、PD 结果文件。
        :param score_name: 分数字段名称。
        :param target_name: 目标变量名称。
        :param bins: 分数分析分箱数量。
        :param border: 分数的最大最小边界， [min, max]。
        :param keys: 用于多样本中区分不同样本的lift。
        :param logger: 日志记录器。
        :param verbose: 是否打印日志。
        """
        self.scores = scores
        self.score_name = score_name
        self.target_name = target_name
        self.bins = bins
        self.border = border
        self.keys = keys
        self.logger = logger
        self.verbose = verbose
        self.kwargs = kwargs
        self.lift_columns = ['total', 'bad', 'bad_rate', 'cum_bad', 'ks', 'lift',]
        self.psi_columns = ['psi', 'psi_val']

        self._validate_data()
        self._lift_borders()

    def _validate_data(self):
        """"""
        if isinstance(self.scores, DataFrame):
            self.scores = [self.scores]

        for i, score in enumerate(self.scores):
            if self.score_name not in score.columns.tolist():
                raise ValueError(f"score_name {self.score_name} not in score columns, score idx is {i}.")
            if self.target_name not in score.columns.tolist():
                # raise ValueError(f"target_name {self.target_name} not in score columns, score idx is {i}.")
                self._logger(msg=f"[{__class__.__name__}] target_name {self.target_name} not in score columns, score idx is {i}.", color='red')
        if self.keys is None:
            self.keys = [f"score-{i}" for i, score in enumerate(self.scores)]

    def _lift_borders(self):
        """"""
        if self.border is None:
            self.border = (0, 100)

        if isinstance(self.bins, int):
            self.bins = np.linspace(*self.border, self.bins + 1)

    def lift(self, score: DataFrame) -> DataFrame:
        """ 计算模型结果的lift """
        return lift(score=score, target_name=self.target_name, score_name=self.score_name, bins=self.bins)

    def score_lifts(self) -> DataFrame:
        """"""
        self.lifts = []
        keys = []
        for key, score in zip(self.keys, self.scores):
            if self.target_name not in score.columns.tolist():
                self._logger(msg=f"[{__class__.__name__}] Skip Evaluate. Message: target_name {self.target_name} not in score columns", color='red')
            else:
                df_lift = self.lift(score)[self.lift_columns]
                self.lifts.append(df_lift)
                keys.append(key)
        self.df_lifts = pd.concat(self.lifts, keys=keys, join="outer", axis=1)
        return self.df_lifts

    def score_psi(
            self,
            score_idx: Optional[Tuple[int, int]] = None,
    ) -> DataFrame:
        """"""
        if len(self.scores) < 2:
            raise ValueError("At least two scores are required to calculate PSI")

        if len(self.lifts) == 0:
            raise ValueError("Lifts have not been calculated yet")

        if score_idx is None:
            score_idx = (0, 1)
        origin_dist = self.lifts[score_idx[0]].total
        current_dist = self.lifts[score_idx[1]].total
        df_psi_data = pd.DataFrame({
            "origin": origin_dist,
            "current": current_dist
        })
        df_psi = psi(df_psi_data, ex_name='origin', ac_name='current')
        return df_psi[self.psi_columns]

    def add_psi(
            self,
            score_idx: Optional[Tuple[int, int]] = None,
    ) -> Union[DataFrame, None]:
        """"""
        if len(self.scores) < 2 or len(self.lifts) < 2:
            self._logger(msg=f"[{__class__.__name__}] At least two scores are required to calculate PSI.", color="red")
            # raise ValueError("At least two scores are required to calculate PSI")
            return None
        else:
            if "psi" not in self.df_lifts.columns:
                df_psi = self.score_psi(score_idx=score_idx)
                self.df_lifts = pd.concat([self.df_lifts, df_psi], axis=1)
            else:
                self._logger(msg=f"[{__class__.__name__}] PSI already exists in {self.df_lifts.columns}", color="blue")
            return self.df_lifts

    def score_evaluate(
            self,
            score_idx: Optional[Tuple[int, int]] = None,
    ):
        """"""
        self.score_lifts()
        self.add_psi(score_idx=score_idx)


class ModelEvaluate(ScoreEvaluate):
    """"""
    model: Any
    weight: DataFrame

    def __init__(
            self,
            samples: List[DataFrame],
            columns: Optional[List[str]] = None,
            train_sample_idx: Optional[int] = 0,
            id_name: str = "uuid",
            target_name: str = "target",
            score_range: Optional[Tuple[int, int]] = None,
            decimal: Optional[int] = 2,
            threshold: Optional[float] = None,
            bins: Optional[Union[int, List[float]]] = 20,
            border: Optional[Tuple[Union[int, float], Union[int, float]]] = None,
            logger: Optional[Callable] = None,
            verbose: Optional[bool] = False,
            **kwargs: Any,
    ):
        """
        :param samples: List of DataFrame
        :param columns: List of columns to evaluate
        :param train_sample_idx: Index of the training sample
        :param id_name: Name of the id column
        :param target_name: Name of the target column
        :param score_range: Range of the score
        :param decimal: Number of decimal places
        :param threshold: Threshold for bad predictions
        :param bins: Number of bins for score
        :param border: Border for score
        :param logger: Logger
        :param verbose: Verbose mode
        """
        self.samples = samples
        self.columns = columns
        self.train_sample_idx = train_sample_idx
        self.id_name = id_name
        self.target_name = target_name
        self.score_range = score_range
        self.decimal = decimal
        self.logger = logger
        self.verbose = verbose
        self.threshold = self.set_threshold(threshold=threshold)
        self.columns = self.set_columns(columns=columns)
        self.scores_metrics = []
        self.model_fit()
        scores = self.model_inference()
        super().__init__(
            scores=scores, target_name=target_name, logger=logger, verbose=verbose,
            bins=bins, border=border, **kwargs,)

    def set_columns(self, columns: Optional[List[str]] = None) -> List[str]:
        """"""
        if columns is None:
            columns = self.samples[self.train_sample_idx].columns.tolist()
        for col in [self.id_name, self.target_name]:
            if col in columns:
                columns.remove(col)
        return columns

    def set_threshold(self, threshold: Optional[float] = None) -> float:
        if threshold is None:
            threshold = self.samples[self.train_sample_idx][self.target_name].mean()
        return threshold

    def model_score(
            self,
            sample: DataFrame,
    ) -> Series:
        """  """
        score = sample[self.columns].dot(self.weight.weight) / 100.
        if self.score_range:
            a, b = self.score_range
            score = NumRange(a, b, score.min(), score.max(), score)
        return score.round(self.decimal)

    def model_inference(self) -> List[DataFrame]:
        """"""
        scores = []
        for sample in self.samples:
            df_sample = sm.add_constant(sample[self.columns])
            proba = self.model.predict(df_sample)
            score_data = {
                "uuid": sample[self.id_name],
                "proba": proba,
                "score": self.model_score(sample),
                "prediction": (proba > self.threshold).astype(int),
            }
            if self.target_name in sample.columns.tolist():
                score_data.update({"target": sample[self.target_name],})
            scores.append(pd.DataFrame(score_data))
        return scores

    def model_fit(self):
        """"""
        x_train = self.samples[self.train_sample_idx][self.columns]
        y_train = self.samples[self.train_sample_idx][self.target_name]
        self.model = StatsLogit(X=x_train, y=y_train)
        self._logger(msg=f"{str(self.model.summary())}\n", color='green')
        self.weight = pd.DataFrame(self.model.params.drop('const'), columns=['param'])
        self.weight['weight'] = self.weight['param'] * 100 / self.weight['param'].sum()

    def model_evaluate(self):
        """"""
        for i, score in enumerate(self.scores):
            if self.target_name in score.columns.tolist():
                binary_metrics = BinaryMetrics(target=score.target, proba=score.proba, prediction=score.prediction)
                score_metrics = binary_metrics.evaluate()
                self.scores_metrics.append(score_metrics)
                metrics_info = f"AUC = {score_metrics.auc:.2f}, KS = {score_metrics.ks:.2f}, Lift = {score_metrics.lift:.2f}, recall = {score_metrics.recall:.2f}, precision = {score_metrics.precision:.2f}, f1-score = {score_metrics.f1_score:.2f}\n"
                self._logger(msg=f"[{__class__.__name__}] Sample ID {i} Metrics: {metrics_info}", color='blue')
            else:
                self._logger(msg=f"[{__class__.__name__}] Sample ID {i} Metrics: No target in score", color='red')

    def evaluate(
            self,
            psi_score_idx: Optional[Tuple[int, int]] = None
    ):
        """"""
        self.model_evaluate()
        self.score_evaluate(score_idx=psi_score_idx)



import statsmodels.api as sm
from pyspark.sql.types import *
from pyspark.sql import functions as F


__all__ = [
    'StatsLogit',
    'NumRange',
    'mode',
    'Div',
]


def StatsLogit(X, y, intercept=True):
    """
    描述： statsmodels 逻辑回归
    :param X:
    :param y:
    :param intercept:
    :return:
    """
    if intercept:
        X = sm.add_constant(X, has_constant='add')
    model = sm.Logit(y, X).fit(disp=False)
    return model


def NumRange(a, b, X_min, X_max, X):
    """
    描述：将数值转换为任意值域
    :param a: 期望最小值
    :param b: 期望最大值
    :param X_min: 数据最小值
    :param X_max: 数据最大值
    :param X: 数据
    :return:
    """
    return a + ((b - a) / (X_max - X_min)) * (X - X_min)


def mode(series):
    """
    描述：
    :param series:
    :return:
    """
    try:
        return float(series.mode())
    except:
        return -9999.


class Div:
    @staticmethod
    def divSpark(df, div_name, a, b, else_value=-9999.):
        """
        描述：Spark 计算 Div 指标
        :param df:
        :param div_name:
        :param a:
        :param b:
        :param else_value:
        :return:
        """
        df = df.withColumn(div_name, F.when(F.col(b) != 0., F.col(a) / F.col(b)).otherwise(else_value))
        return df

    @staticmethod
    def divPandas(df, div_name, a, b, else_value=-9999.):
        """
        描述：Pandas 计算 Div 指标
        :param df:
        :param div_name:
        :param a:
        :param b:
        :param else_value:
        :return:
        """
        df[div_name] = df[a] / df[b].replace(0, 1e-6)
        return df

    @staticmethod
    def div(a, b, else_value=-9999.):
        if b != 0:
            return a / b
        else:
            return else_value



try:
    from pyspark.sql.window import Window
    from pyspark.sql import functions as F
    from pyspark.sql.types import StringType, DoubleType
    from pyspark.sql import DataFrame as TypeSparkDataFrame
except ImportError:
    print("pyspark not installed, please install it first. < pip install pyspark >")
from typing import List, Optional
from pandas import DataFrame as TypePandasDataFrame


__all__ = [
    "pct_rank",
]


def pct_rank(
        df: TypePandasDataFrame,
        column_name,
        index_name: str = None,
        partition_by: Optional[List[str]] = None,
) -> TypePandasDataFrame:
    """
    描述：返回依据行业作为窗口的行业排名 DataFrame
    :param df: 原始数据
    :param column_name: 新指标名称
    :param index_name: 原指标名称
    :param partition_by: 窗口字段
    :return: spark DataFrame
    """
    if partition_by is None:
        partition_by = ['load_date', 'industry']

    if index_name is None:
        index_name = column_name
    # window of industry rank
    wind_indu = Window.partitionBy(*partition_by).orderBy(F.desc(column_name))
    df = df.withColumn(f'rank_indu_{index_name}', F.percent_rank().over(wind_indu))
    return df





class Apriori:
    def __init__(self, dataSet, minSupport=.5, ruleSupport=.7, is_show_rule=False):
        self.dataSet = dataSet
        self.minSupport = minSupport
        self.ruleSupport = ruleSupport
        self.is_show_rule = is_show_rule
        # 计算支持项，全部支持项与最小支持项
        self.L, self.supportData = self.apriori()

        # 计算关联规则
        self.rules = self.generateRules(self.L, self.supportData, self.ruleSupport)


    # C1 是大小为1的所有候选项集的集合
    def createC1(self, ):
        C1 = []
        for transaction in self.dataSet:
            for item in transaction:
                if not [item] in C1:
                    C1.append([item])  # store all the item unrepeatly

        C1.sort()
        # return map(frozenset, C1); frozen set, user can't change it.
        return list(map(frozenset, C1))


    def scanD(self, D, Ck):
        # 这是一个计数器，dict类型
        ssCnt = {}
        for tid in D:
            for can in Ck:
                # can 是否是 tid 的子集
                if can.issubset(tid):
                    # if not ssCnt.has_key(can):
                    if not can in ssCnt:
                        ssCnt[can] = 1
                    else:
                        ssCnt[can] += 1
        numItems = float(len(D))
        retList = []
        supportData = {}
        for key in ssCnt:
            support = ssCnt[key] / numItems  # compute support
            if support >= self.minSupport:
                retList.insert(0, key)
            supportData[key] = support
        return retList, supportData


    # apriori
    def apriori(self,):
        # 单个元素
        C1 = self.createC1()
        # 不重复的数据集
        D = list(map(set, self.dataSet))  # python3
        # 生成最小支持度的L1
        L1, supportData = self.scanD(D, C1)  # 单项最小支持度判断 0.5，生成L1
        L = [L1]
        k = 2

        while (len(L[k - 2]) > 0):  # 创建包含更大项集的更大列表,直到下一个大的项集为空
            Ck = self.aprioriGen(L[k - 2], k)  # Ck
            Lk, supK = self.scanD(D, Ck)  # get Lk
            supportData.update(supK)
            L.append(Lk)
            k += 1  # 继续向上合并 生成项集个数更多的
        return L, supportData


    # total apriori
    def aprioriGen(self, Lk, k):  # 组合，向上合并
        # creates Ck 参数：频繁项集列表 Lk 与项集元素个数 k
        retList = []
        lenLk = len(Lk)
        for i in range(lenLk):
            for j in range(i + 1, lenLk):  # 两两组合遍历
                L1 = list(Lk[i])[:k - 2]
                L2 = list(Lk[j])[:k - 2]
                L1.sort()
                L2.sort()
                if L1 == L2:  # 若两个集合的前k-2个项相同时,则将两个集合合并
                    retList.append(Lk[i] | Lk[j])  # set union
        return retList


    # 生成关联规则
    def generateRules(self, L, supportData, minConf=0.7):
        '''
        L: 频繁项集列表
        supportData: 包含那些频繁项集支持数据的字典
        minConf: 最小可信度阈值
        '''

        # 存储所有的关联规则
        bigRuleList = []

        # 只获取有两个或者更多集合的项目，从1,即第二个元素开始，L[0]是单个元素的
        for i in range(1, len(L)):
            # 两个及以上的才可能有关联一说，单个元素的项集不存在关联问题
            for freqSet in L[i]:
                H1 = [frozenset([item]) for item in freqSet]
                # 该函数遍历L中的每一个频繁项集并对每个频繁项集创建只包含单个元素集合的列表H1
                if (i > 1):
                    # 如果频繁项集元素数目超过2,那么会考虑对它做进一步的合并
                    self.rulesFromConseq(freqSet, H1, supportData, bigRuleList, minConf)
                else:  # 第一层时，后件数为1
                    self.calcConf(freqSet, H1, supportData, bigRuleList, minConf)  # 调用函数2
        return bigRuleList


    # 生成候选规则集合：计算规则的可信度以及找到满足最小可信度要求的规则
    def calcConf(self, freqSet, H, supportData, brl, minConf=0.7):
        # 针对项集中只有两个元素时，计算可信度
        prunedH = []  # 返回一个满足最小可信度要求的规则列表
        for conseq in H:  # 后件，遍历 H中的所有项集并计算它们的可信度值
            conf = supportData[freqSet] / supportData[freqSet - conseq]  # 可信度计算，结合支持度数据
            if conf >= minConf:
                if self.is_show_rule:
                    print(freqSet - conseq, '-->', conseq, 'conf:', conf)
                # 如果某条规则满足最小可信度值,那么将这些规则输出到屏幕显示
                brl.append((freqSet - conseq, conseq, conf))  # 添加到规则里，brl 是前面通过检查的 bigRuleList
                prunedH.append(conseq)  # 同样需要放入列表到后面检查
        return prunedH

    # 合并
    def rulesFromConseq(self, freqSet, H, supportData, brl, minConf=0.7):
        # 参数:一个是频繁项集,另一个是可以出现在规则右部的元素列表 H
        m = len(H[0])
        if (len(freqSet) > (m + 1)):  # 频繁项集元素数目大于单个集合的元素数
            Hmp1 = self.aprioriGen(H, m + 1)  # 存在不同顺序、元素相同的集合，合并具有相同部分的集合
            Hmp1 = self.calcConf(freqSet, Hmp1, supportData, brl, minConf)  # 计算可信度
            if (len(Hmp1) > 1):  # 满足最小可信度要求的规则列表多于1,则递归
                self.rulesFromConseq(freqSet, Hmp1, supportData, brl, minConf)





from .utils import *


class MasterScale(BetaDistribution):

    def __init__(self, current_rank_rate: list, target_level: int, samples: int = 1e6):
        """
        初始化
        :param current_rank_rate: 目前评级分布，即目前每一个评级的占比
        :param target_level: 目标细分评级等级数
        :param samples: 随机数模拟 Beta 分布样本数
        """
        super(MasterScale).__init__()
        self.samples = samples
        self.current_rank_rate = current_rank_rate
        self.target_level = target_level
        self.grades = np.arange(1, self.target_level + 1e-6)

        # 计算目前评级分布的 Beta 分布参数
        self.__cal_beta_params()

    def __generate_random(self, ):
        """
        生成模拟Beta分布的随机数样本
            1. [0, 1] 内划分 current_rank_rate 个间断
            2. 在间断点内按照相应比例生成均匀分布随机数
        :return: 随机数序列
        """
        arange = np.arange(0, 1 + 1e-6, 1 / len(self.current_rank_rate))
        random_nums = np.array([])
        for i, level_rate in enumerate(self.current_rank_rate):
            level_random = np.random.uniform(
                low=arange[i],
                high=arange[i + 1],
                size=int(level_rate * self.samples)
            )
            random_nums = np.hstack((random_nums, level_random))
        return random_nums

    def __cal_beta_params(self, ):
        """
        计算 Beta 分布的分布参数
        :return: None
        """
        # 生成模拟随机数
        random_nums = self.__generate_random()
        # 计算分布参数
        self.Alpha, self.Beta, lower, upper = beta.fit(random_nums, floc=0, fscale=1)

    def cal_quantile(self, ranking: np.array, target_grade: int):
        """
        计算真实 PD 或 Score 排序下的新主标尺分布
        :param ranking: 真实排序 PD or Score
        :param target_grade: 目标主标尺总等级
        :return:
        """
        target_grade_rate, CDF = self.grade_rate(target_grade)
        # 计算 Beta 分布下分位点的 PD 或 Score 值
        quantile = [np.quantile(ranking, cdf) for cdf in CDF]
        # 新主标尺评级等级
        levels = np.empty_like(ranking)
        # 评级赋值
        for level in range(target_grade):
            if level == 0:
                levels[ranking <= quantile[level]] = level
            else:
                levels[np.logical_and(ranking <= quantile[level], ranking > quantile[level - 1])] = level

    def fit_grade_df(self, grade: np.array, df: np.array, log_df=True):
        """
        线性拟合
        :param grade: 等级
        :param df: 等级对应的 Default rate
        :param log_df: 是否对 Default rate 取对数，默认为 True
        :return: None
        """
        import statsmodels.api as sm
        X = sm.add_constant(grade)
        if log_df:
            self.OLS_model = sm.OLS(np.log(df), X).fit(disp=False)
        else:
            self.OLS_model = sm.OLS(df, X).fit(disp=False)
        self.intercept, self.coef = self.OLS_model.params.tolist()

        # 添加常数项
        pre_X = sm.add_constant(self.grades)
        self.grade_df = np.e**(self.OLS_model.predict(pre_X, 1))

    def confirm_master_scale(self, grade_df=None):
        """
        确认评级等级的PD，即评级等级PD的上下界
        :return: 结果在 class.master_scale
        """
        if grade_df:
            grade_df_mean = (grade_df[1:] + grade_df[:-1]) / 2
        else:
            grade_df_mean = (self.grade_df[1:] + self.grade_df[:-1]) / 2
        grade_df_min = np.hstack((np.array([0]), grade_df_mean))
        grade_df_max = np.hstack((grade_df_mean, np.array([1])))
        self.master_scale = pd.DataFrame({
            'grades': self.grades,
            'grade_df_min': grade_df_min,
            'grade_df': self.grade_df,
            'grade_df_max': grade_df_max,
        })

    def plot_grade_rate(self, grades: int):
        """
        根据 Beta 分布绘制 hist
        :param grades: hist nums
        :return: plt.show()
        """
        target_grade_rate, cdf = self.grade_rate(grades)
        plt.figure(figsize=(8, 4), dpi=200)
        plt.bar(range(1, len(target_grade_rate) + 1), target_grade_rate)
        for x, y in zip(range(1, len(target_grade_rate) + 1), target_grade_rate):
            plt.text(x - 0.5, y + 0.001, f'{y * 100: .1f}%', fontsize=6)
        return plt.show()



import pandas as pd
from pandas import DataFrame
from typing import Any, List, Callable, Literal, Optional
from ..utils import dispatch_tqdm, batches
from ..utils.logger import LoggerMixin


__all__ = [
    'WOEScore'
]


class WOEScore(LoggerMixin):
    """"""
    def __init__(
            self,
            df: DataFrame,
            df_woe: DataFrame,
            id_name: Optional[str] = 'uuid',
            target_name: Optional[str] = 'target',
            dtype: Optional[str] = 'float',
            batch_size: Optional[int] = 32,
            error='error',
            logger: Optional[Callable] = None,
            verbose: Optional[bool] = True,
            *args: Any,
            **kwargs: Any

    ):
        """
        :param df: 原始数据
        :param df_woe: woe编码后的数据
        :param id_name: uuid
        :param target_name: 目标变量名称，默认为 target
        :param dtype: 数据类型，默认为float
        :param batch_size: 批量大小，默认为32
        :param error: 错误处理方式，默认为error
        :param logger: 日志记录器，默认为None
        :param verbose: 是否显示进度条，默认为True
        """
        self.df = df
        self.df_woe = df_woe
        self.id_name = id_name
        self.target_name = target_name
        self.dtype = dtype
        self.batch_size = batch_size
        self.error = error
        self.logger = logger
        self.verbose = verbose
        self.args = args
        self.kwargs = kwargs

        self._validate_data()
        self._prepare_data()
        self._set_private_data()

    def _set_private_data(self):
        """"""
        data_columns = self.df.columns.tolist()
        if self.id_name in data_columns and self.target_name in data_columns:
            self.df_private = self.df[[self.id_name, self.target_name]].copy()
        elif self.id_name in data_columns:
            self.df_private = self.df[[self.id_name]].copy()
        else:
            raise ValueError('id_name and target_name must be in data_columns')

    def _validate_woe_data(self):
        """"""
        _validate_mark = True
        for group_name, group in self.df_woe.groupby('var_name'):
            if group.min_bin.isna().sum() == 0 and self.df[group_name].isna().sum() > 0:
                self._logger(msg=f"[{__class__.__name__}] WOE does not contain nan bin but there are nan values in the data <{group_name}>, please check the data: {group_name}.", color="magenta")
                _validate_mark = False
        return _validate_mark

    def _validate_data(self):
        """
        """
        if self.id_name not in self.df.columns.tolist():
            raise ValueError(f'uuid {self.id_name} not in columns, please check your data.')

        if len(self.df) != len(self.df[self.id_name].unique()):
            raise ValueError('uuid is not unique, please check your data.')

        # if self.target_name not in self.df.columns.tolist():
        #     raise ValueError(f'target {self.target_name} not in columns, please check your data.')

        if not self._validate_woe_data():
            raise ValueError('WOE data is not valid, please check your data.')

    def _prepare_data(self):
        """"""
        self.df.reset_index(inplace=True, drop=True)

    def _trans_columns(self, columns: Optional[List[str]] = None):
        """"""
        if columns is None:
            columns = self.df.columns.tolist()
        for _drop in [self.id_name, self.target_name]:
            if _drop in columns:
                columns.remove(_drop)
        woe_columns = self.df_woe.var_name.unique().tolist()
        columns = list(set(woe_columns) & set(columns))
        return columns

    def wide_to_long(self, df_wide: DataFrame, columns: Optional[List[str]] = None) -> DataFrame:
        """ 宽表转为长表，用于WOE分数转换 """
        df_long = df_wide[columns].add_prefix('values_').join(self.df_private[[self.id_name]])
        df_long = pd.wide_to_long(
            df_long, stubnames='values',
            i=self.id_name, j='var_name', sep='_', suffix='\w+'
        )
        df_long.reset_index(inplace=True)
        df_long.reset_index(inplace=True, drop=True)
        df_long['values'] = df_long['values'].astype(self.dtype)
        return df_long

    def _match_miss_score(self, df_merge_score: DataFrame) -> DataFrame:
        """"""
        cond_miss = df_merge_score[['values', 'min_bin', 'max_bin']].isna().all(axis=1)
        df_miss_score = df_merge_score[cond_miss].copy()
        return df_miss_score

    def _match_normal_score(self, df_merge_score: DataFrame) -> DataFrame:
        """"""
        cond_normal = df_merge_score[['values', 'min_bin', 'max_bin']].isna().any(axis=1)
        df_score = df_merge_score[~cond_normal].copy()

        cond_eq = (df_score['values'] == df_score.min_bin) & (df_score['values'] == df_score.max_bin)
        cond_mid = (df_score['values'] > df_score.min_bin) & (df_score['values'] <= df_score.max_bin)
        cond_score = cond_mid | cond_eq

        df_score = df_score[cond_score]
        return df_score

    def match_woe(self, df_long: DataFrame, df_woe: DataFrame) -> DataFrame:
        """ """
        df_merge_score = df_long.merge(df_woe, on='var_name', how='left')
        df_normal_score = self._match_normal_score(df_merge_score)
        df_miss_score = self._match_miss_score(df_merge_score)
        df_score = pd.concat([df_normal_score, df_miss_score], axis=0, ignore_index=True)

        if self.error == 'error':
            if df_score.shape[0] != df_long.shape[0]:
                raise ValueError(f'sample shape {df_score.shape[0]} not match origin dataset shape {df_long.shape[0]}.')
        elif self.error == 'ignore':
            if df_score.shape[0] > df_long.shape[0]:
                self._logger(msg=f'Mutil Bins! {df_score.shape}, {df_long.shape}', color="red")
            elif df_score.shape[0] < df_long.shape[0]:
                self._logger(msg=f'Bins Margin Warn! {df_score.shape}, {df_long.shape}', color="red")
        return df_score

    def trans_woe_score(
            self,
            columns: Optional[List[str]] = None,
            values: Optional[Literal['woe', 'score']] = "score",
            add_target: Optional[bool] = True,
            batch_info: Optional[str] = None,
    ) -> DataFrame:
        """ 两种种WOE映射值 """
        columns = self._trans_columns(columns=columns)
        df_select = self.df[columns].copy()
        df_woe_select = self.df_woe[self.df_woe.var_name.isin(columns)].copy()

        if batch_info:
            logger_name = f"[{__class__.__name__}] {batch_info}"
        else:
            logger_name = f"[{__class__.__name__}]"

        self._logger(msg=f'{logger_name} Wide to Long ...', color="green")
        df_long = self.wide_to_long(df_select, columns=columns)
        self._logger(msg=f'{logger_name} Match WOE ...', color="green")
        df_score_long = self.match_woe(df_long=df_long, df_woe=df_woe_select)
        self._logger(msg=f'{logger_name} Match Over ...', color="green")

        df_score_wide = df_score_long.pivot(index=self.id_name, columns='var_name', values=values)
        if add_target:
            df_score_wide = df_score_wide.join(self.df_private.set_index(self.id_name), on=self.id_name)
        df_score_wide.sort_index(inplace=True)
        return df_score_wide

    def batch_run(
            self,
            columns: Optional[List[str]] = None,
            values: Optional[Literal['woe', 'score']] = "score",
            id_index: Optional[bool] = False,
    ):
        """ """
        columns = self._trans_columns(columns=columns)
        self._logger(msg=f"Start Trans WOE Score, total columns number is {len(columns)} ...", color="green")
        batch_list = [batch for batch in batches(lst=columns, batch_size=self.batch_size)]
        _batches = []
        batch_idx = 1
        for batch_columns in dispatch_tqdm(batch_list, desc=f'[{__class__.__name__}] WOE Batch'):
            df_batch = self.trans_woe_score(
                columns=batch_columns, values=values, add_target=False, batch_info=f'[{batch_idx} / {len(batch_list)}]')
            _batches.append(df_batch)
            batch_idx += 1
        df_score = pd.concat(_batches, axis=1, join="outer",)
        df_score = df_score.join(self.df_private.set_index(self.id_name), on=self.id_name)
        if not id_index:
            df_score.reset_index(inplace=True)
        return df_score



from hashlib import md5
try:
    from pyspark.sql import functions as F
    from pyspark.sql.types import StringType
except ImportError:
    print("pyspark not installed, please install it first. < pip install pyspark >")


__all__ = [
    "md5", "md5_udf",
]


def md5(x):
    """MD5生成器"""
    return md5(str(x).encode("utf8")).hexdigest()


md5_udf = F.udf(f=md5, returnType=StringType())



import os
from typing import Optional
try:
    import pyspark
    from pyspark.sql import SparkSession
    from pyspark import SparkService
except ImportError:
    print("pyspark not installed, please install it first. < pip install pyspark >")


__all__ = [
    "enable_arrow",
    "get_spark",
    "kill_app",
]


def enable_arrow(spark: SparkSession, verbose: Optional[bool] = False) -> None:
    """
    描述：启用 Arrow
    """
    spark.conf.set("spark.sql.execution.arrow.enabled", "true")
    if verbose:
        print(f'Enable Arrow: {spark.conf.get("spark.sql.execution.arrow.enabled")}')


def get_spark(
        app_name: Optional[str] = 'test',
        master: Optional[str] = 'yarn',
        executors: Optional[int] = 2
) -> SparkSession:
    """

    :param app_name:
    :param master: 'yarn', 'local[*]'
    :param executors:
    :return:
    """
    if hasattr(pyspark, "SparkService"):
        extra_param = {
            'spark.app.name': app_name,
            'spark.driver.cores': 2,
            'spark.sql.execution.arrow.pyspark.enabled': 'true',
        }
        spark = SparkService.get_yarn_spark(
            executor_instances=executors,
            per_executor_mem='2g',
            driver_mem='1g',
            extra_param=extra_param,
        )
    elif hasattr(pyspark.sql, 'SparkSession'):
        from pyspark.sql import SparkSession
        builder = SparkSession.builder \
            .appName(app_name) \
            .enableHiveSupport() \
            .master(master) \
            .config("spark.dynamicAllocation.maxExecutors", f"{executors}")
        spark = builder.getOrCreate()
    else:
        raise ValueError("Not support for Spark, please check your env.")
    return spark


def kill_app(app_id: str) -> None:
    """
    描述：kill spark application

    :param app_id: spark app id
    :return:
    """
    info = os.popen(f'yarn application --kill {app_id}').readlines()
    print('\n'.join(info))



import random
import hashlib
import numpy as np
import pandas as pd
from typing import *
from .fake_data_config import configs
from multiprocessing import Pool

__all__ = [
    'map_data_type',
    'gen_data',
    'generate_credit_data',
]

np.random.seed(1024)
map_data_type = {
    'uniform': np.random.uniform,
    'randint': np.random.randint,
}

#### TODO: 同一个用户多个时间的数据。
####

def gen_data(
        num_samples: int,
        low: Union[float, int],
        high: Union[float, int],
        data_type: str,
) -> np.array:
    """"""
    data_generator = map_data_type.get(data_type)
    data = data_generator(low, high, size=num_samples)
    return data


def generate_credit_data(
        num_samples: int,
        configs: List[Dict],
        target_ratio: float = 0.2,
        seed: int = 1024,
) -> pd.DataFrame:
    """"""
    num_good = int(num_samples * (1 - target_ratio))
    num_bad = num_samples - num_good

    data = dict()
    for config in configs:
        index_name = config.get('index_name')
        bad_range, good_range = config.get('index_range')
        index_type = config.get('index_type')
        index_bad_val = gen_data(num_bad, bad_range[0], bad_range[1], data_type=index_type, )
        index_good_val = gen_data(num_good, good_range[0], good_range[1], data_type=index_type, )
        index_val = np.concatenate((index_good_val, index_bad_val))
        data[index_name] = index_val
    df = pd.DataFrame(data)
    df['target'] = 0
    df.loc[num_good:, 'target'] = 1
    df = df.sample(frac=1, random_state=seed).reset_index(drop=True)
    df['uuid'] = df.index.map(calculate_md5)
    return df


def random_date(min_date, max_date):
    """"""
    min_date = int(min_date)
    max_date = int(max_date)

    min_datetime = pd.to_datetime(str(min_date), format='%Y%m%d')
    max_datetime = pd.to_datetime(str(max_date), format='%Y%m%d')
    random_datetime = random.choice(pd.date_range(min_datetime, max_datetime))
    random_date_str = random_datetime.strftime('%Y%m%d')
    return random_date_str


def calculate_md5(string):
    """
    desc: 生成md5编码
    :param string:
    :return:
    """
    md5_hash = hashlib.md5()
    md5_hash.update(string.encode('utf-8'))
    md5_encoded = md5_hash.hexdigest()
    return md5_encoded





import xlsxwriter
import seaborn as sns


class BeautyExcel:

    def __init__(self, df=None, dfs=None, path=None):
        """"""
        self.df = df
        self.dfs = dfs
        self.path = path
        # self.max_row, self.max_col = self.df.shape

        self.workbook = xlsxwriter.Workbook(self.path, {'nan_inf_to_errors': True})

    def get_pos(self, df, pos=None, start_row=0, start_col=0, interval=0, how='left'):
        """
        """
        if pos is not None and how == 'left':
            start_row, _, _, start_col = pos
            start_col += interval
        if pos is not None and how == 'bottom':
            _, start_col, start_row, _ = pos
            start_row += interval

        rows, cols = df.shape
        end_row = start_row + rows
        end_col = start_col + cols - 1
        return (start_row, start_col, end_row, end_col)

    def get_sheet(self, sheet_name):
        """"""
        try:
            return self.workbook.add_worksheet(sheet_name)
        except:
            return self.workbook.get_worksheet_by_name(sheet_name)

    def set_columns(self, df=None, columns=None):
        """"""
        if df is not None:
            columns = df.columns
        return [{'header': column} for column in columns]

    def set_rows_width(self, sheet_name, pos, width):
        """
        """
        self.get_sheet(sheet_name).set_row(pos, width)

    def set_columns_width(self, sheet_name, pos, width):
        """
        """
        if isinstance(pos, tuple):
            self.get_sheet(sheet_name).set_column(*pos, width)
        if isinstance(pos, str):
            self.get_sheet(sheet_name).set_column(pos, width)

    def set_header(self, sheet_name, text):
        """
        :text:
            - &[L, C, R]   位置
            - &[D, T]      时间
            - &[U, E]      下划线
            - &[], &G      图片占位符
            - &[F]         文件名
        """
        self.get_sheet(sheet_name).set_header(text)

    def set_data_bar(self, sheet_name, pos=None, column_name=None):
        """
        """
        if column_name is not None:
            column_id = self.columns.index(column_name)
            self.get_sheet(sheet_name).conditional_format(
                1, column_id, self.max_row, column_id, {'type': 'data_bar'})
        else:
            self.get_sheet(sheet_name).conditional_format(
                *pos, {'type': 'data_bar'})

    def add_text(self, ):
        """"""

    def gen_cell_format(self, palette='Paired', n_colors=None, desat=None, ):
        """
        """
        colormap = list(sns.color_palette(palette=palette, n_colors=n_colors, desat=desat).as_hex())
        cell_formats = [self.workbook.add_format({'bg_color': color}) for color in colormap]
        return cell_formats

    def set_condition_fromat(self, sheet_name, pos, criteria, value, cell_format, cell_type='cell', ):
        """
        描述： 添加条件

        :param type: `cell`, `data_bar`, `text`
        :param criteria: `>`, `=`, `>=`, `<=`, `containing`
        :param value: value
        :param cell_format: cell_format
        """
        param = {
            'type': cell_type,
            'criteria': criteria,
            'value': value,
            'format': cell_format,
        }

        self.get_sheet(sheet_name).conditional_format(
            *pos, param)

    def add_url(self, sheet_name, pos, url, string=None):
        """
        描述：
            - eg: internal:Sheet2!A1
            - eg: internal:Sheet2!A1
        """
        if string is None:
            string = url
        if isinstance(pos, tuple):
            self.get_sheet(sheet_name).write_url(*pos, url, string=string)
        if isinstance(pos, str):
            self.get_sheet(sheet_name).write_url(pos, url, string=string)

    def add_img(
            self, sheet_name, row, col, filename, shift=(0, 0), scale=(1, 1),
            object_position=2, image_data=None, url=None, description=None,
            decorative=False):
        """"""

        x_offset, y_offset = shift
        x_scale, y_scale = scale

        param = {
            'x_offset': x_offset,
            'y_offset': y_offset,
            'x_scale': x_scale,
            'y_scale': y_scale,
            'object_position': object_position,
            'image_data': image_data,
            'url': url,
            'description': description,
            'decorative': decorative,
        }

        self.get_sheet(sheet_name).insert_image(row, col, filename, param)

    def add_table(self, df=None, sheet_name=None, pos=None, param=None, style='Table Style Light 8'):
        """

        """
        if sheet_name is None:
            sheet_name = 'Sheet'

        if pos is None:
            start_row, start_col = 0, 0
            end_row, end_col = df.shape
            end_col -= 1
        else:
            start_row, start_col, end_row, end_col = pos

        if param is None:
            param = {
                'data': df.values.tolist(),
                'columns': self.set_columns(df=df),
                'style': style,
            }

        self.get_sheet(sheet_name).add_table(
            start_row, start_col, end_row, end_col, param)

    def group(self, df, sheet_name, by, sort_by=None, ascending=False, pos=(0, 0), style=None, interval=1):
        """
        描述：分组展示统计项

        参数：
        :param style:
            - Table Style Light 1-21
            - Table Style Medium 1-28
            - Table Style Dark 1-11
        """
        param = {
            'data': None,
            'columns': self.set_columns(df=df),
            'style': style,
        }
        start_row, start_col = pos

        for group_name, df_group in df.groupby(by):
            if sort_by:
                df_group = df_group.sort_values(by=sort_by, ascending=ascending)
            param['data'] = df_group.values.tolist()
            rows, cols = df_group.shape
            end_row, end_col = start_row + rows, start_col + cols - 1
            self.add_table(df_group, sheet_name=sheet_name,
                           pos=(start_row, start_col, end_row, end_col), param=param)
            start_row = start_row + rows + 1 + interval



import pandas as pd
from pandas import DataFrame, Series
from typing import Any, List, Union, Optional
from kivi.woe.base import WOEMixin


__all__ = ["CategoryBins"]


class CategoryBins(WOEMixin):
    """
    类别型变量的分箱
    """
    def __init__(
            self,
            variables: Series,
            target: Series,
            bins: Optional[int] = 5,
            abnormal_vals: Optional[List[Union[str, int, float]]] = None,
            fill_bin: Optional[bool] = True,
            decimal: Optional[int] = 6,
            weight: Optional[Any] = None,
            *args: Any,
            **kwargs: Any,
    ):
        """
        描述：类别型变量分箱分析。

        :param variables: 待分箱变量
        :param target: 目标标签变量
        :param fill_bin: 在各分箱中偶发性会出现 good 或 bad 为 0 的情况，默认 fill_pos 为 True ，为该分箱填充 0.5。

        Example:
            woe = CategoryBins(variables, target)
            woe.fit()
        """
        self.variables = variables
        self.target = target
        self.bins = bins
        self.max_leaf_nodes = bins
        self.fill_bin = fill_bin
        self.abnormal_vals = abnormal_vals
        self.decimal = decimal
        self.args = args
        self.kwargs = kwargs
        self.data_prepare(
            variables=variables, target=target, weight=weight)

    def fit(
            self,
            score: Optional[bool] = True,
            origin_border: Optional[bool] = False,
            order: Optional[bool] = True,
            **kwargs: Any,
    ) -> DataFrame:
        """
        :param score: 是否增加 WOEMixin score。
        :param origin_border: 是否增加 分箱中的最大值与最小值。
        :param order: 是否增加单调性判断。
        :return: DataFrame WOEMixin result.
        """
        bucket = pd.DataFrame({
            'variables': self.df_data.variables,
            'target': self.df_data.target,
            'bucket': self.df_data.variables,
        }).groupby('bucket', as_index=True)

        self.cal_woe_iv(bucket, score=score, origin_border=origin_border, order=order)
        return self.woe



import numpy as np
import pandas as pd
from typing import List, Union, Tuple, Optional
from pandas import DataFrame


__all__ = [
    "BinsMixin",
    "lift",
]


def lift(
        score: DataFrame,
        score_name: Optional[str] = "score",
        target_name: Optional[str] = "target",
        bins: Optional[Union[int, List[Union[int, float]]]] = 10,
) -> DataFrame:
    """ 计算模型结果的lift """
    score["buckets"] = pd.cut(score[score_name], bins=bins, include_lowest=True)
    df_buckets = score.groupby('buckets', observed=False).agg({target_name: ['count', 'sum']})

    df_buckets.columns = ['total', 'bad']
    df_buckets['good'] = df_buckets.total - df_buckets.bad
    df_buckets['bad_rate'] = df_buckets.bad / df_buckets.total

    df_buckets['cum_total'] = df_buckets.total.cumsum() / df_buckets.total.sum()
    df_buckets['cum_bad'] = df_buckets.bad.cumsum() / df_buckets.bad.sum()
    df_buckets['cum_good'] = df_buckets.good.cumsum() / df_buckets.good.sum()
    df_buckets['ks'] = df_buckets.cum_bad - df_buckets.cum_good
    df_buckets['lift'] = df_buckets.cum_bad / df_buckets.cum_total
    return df_buckets


class BinsMixin:

    def Getcutoffpoint(self, bins, data, cut_type='qcut'):
        """
        Des: 计算cutoff point，包括等距分箱，等频分箱
        :param bins: 分箱数
        :param percent: 是否使用等频分箱，默认为不使用False
        :param kwargs: 其他参数
        :return: cutoff point
        """

        if cut_type == 'qcut':
            _, self.cutoffpoint = pd.qcut(
                data, bins, retbins=True, duplicates='drop',
            )
        elif cut_type == 'cut':
            _, self.cutoffpoint = pd.cut(
                data, bins, retbins=True, include_lowest=True, duplicates='drop',
            )

        self.cutoffpoint[0] = -np.inf
        self.cutoffpoint[-1] = np.inf




class Scale:
    """
    数据的归一化方法：
        1. min-max
        2. Z-Score
    :param df: DataFrame
    """
    @staticmethod
    def MinMax(df):
        df = (df - df.min()) / (df.max() - df.min())
        return df

    @staticmethod
    def MeanMinMax(df):
        df = (df - df.mean()) / (df.max() - df.min())
        return df

    @staticmethod
    def ZScore(df):
        df = (df - df.mean()) / df.std()
        return df



try:
    from pyspark.sql import SparkSession
    from pyspark.sql import functions as F
    from pyspark.sql.types import StructField, StringType
    from pyspark.sql.window import Window
    from pyspark.sql import DataFrame as TypeSparkDataFrame
except ImportError:
    print("pyspark not installed, please install it first. < pip install pyspark >")
from typing import Any, List, Optional
from tqdm import tqdm
import pandas as pd
from pandas import DataFrame as TypePandasDataFrame


__all__ = [
    "long_to_wide",
    "unpivot",
    "pandas_to_spark_frame",
    "spark_to_pandas_frame",
    "tqdm"
]


def long_to_wide(
        df_long: TypeSparkDataFrame,
        index: Optional[List[str]] = None,
        columns: Optional[str] = 'index_name',
        values: Optional[str] = 'index_val'
) -> TypeSparkDataFrame:
    """ 描述：对数据表进行 pivot 操作 """
    if index is None:
        index = ['uuid', 'load_date']
    df_wide = df_long.groupBy(index).pivot(columns).agg(
        F.first(values))
    return df_wide


def unpivot(
        df: TypeSparkDataFrame,
        columns: List[str],
        val_type: Optional[Any] = None,
        index_name='uuid',
        feature_name='name',
        feature_value='value',
        orderby=None
) -> TypeSparkDataFrame:
    """描述：对数据表进行反 pivot 操作"""
    if val_type is not None:
        columns_type = [F.col(col).cast(val_type()) for col in columns]
    else:
        columns_type = columns

    stack_query = []
    for col in columns:
        stack_query.append(f"'{col}', `{col}`")

    if isinstance(index_name, str):
        index_name = [index_name]

    df = df.select(*index_name, *columns_type)

    df = df.selectExpr(
        *index_name, f"stack({len(stack_query)}, {', '.join(stack_query)}) as (`{feature_name}`, `{feature_value}`)"
    )

    if orderby:
        df = df.orderBy(orderby)

    return df


def pandas_to_spark_frame(df: TypePandasDataFrame, spark: SparkSession) -> TypeSparkDataFrame:
    """"""
    df = df.where(pd.notnull(df), None)
    return spark.createDataFrame(df)


def spark_to_pandas_frame(
        df: TypeSparkDataFrame,
        interval: Optional[int] = 5e4,
        index_name: Optional[str] = 'index',
        columns: Optional[List[str]] = None,
        drop_index: Optional[bool] = True,
) -> TypeSparkDataFrame:
    """
    转换PySpark到Pandas DataFrame
    interval: 转换间距
    index_name:
    columns: 排序依据列
    drop_index: 返回值是否删除索引index
    """
    if columns is None:
        window = Window().orderBy(df.columns)
    else:
        window = Window().orderBy(columns)
    df = df.withColumn('index', F.row_number().over(window))
    length = df.count()

    df_pd = pd.DataFrame(columns=df.columns)
    for i in tqdm(range(int(length / interval) + 1)):
        lower = i * interval
        upper = (i + 1) * interval
        df_pd = df_pd.append(df.where((df.index < upper) & (df.index >= lower)).toPandas())

    if drop_index:
        return df_pd.drop(index_name, axis=1)
    else:
        return df_pd

